"""
registration.py

Copyright (C) 2014 Sami Koho
All rights reserved.

This software may be modified and distributed under the terms
of the BSD license.  See the LICENSE file for details.

This file contains functions for registration of microscope image
volumes. The methods are based on Insight Toolkit (www.itk.org),
the installation of which is required in order to run the contained
functions

Currently (04/2014) rigid body registration of three-dimensional volumes
has been implemented. Several metrics 1. least-squares 2. viola-wells mutual
information 3. mattes mutual information are supported implemented

"""


import SimpleITK as sitk
import matplotlib.pyplot as plt
from skimage.feature import register_translation
from scipy.ndimage import fourier_shift

import miplib.processing.itk as ops_itk
import miplib.ui.plots.image as show
import miplib.processing.image as imops
from miplib.data.containers.image import Image

import numpy as np

# region OBSERVERS

# PLOTS
# =============================================================================
# Plotting functions for showing the registration progress.


def start_plot():
    global metric_values

    metric_values = []


def end_plot(fixed, moving, transform):
    global metric_values
    plt.subplots(1, 2, figsize=(10, 8))

    # Plot metric values
    plt.subplot(1, 2, 1)
    plt.plot(metric_values, 'r')
    plt.title("Metric values")
    plt.xlabel('Iteration Number', fontsize=12)
    plt.ylabel('Metric Value', fontsize=12)

    # Plot image overlay
    resampled = ops_itk.resample_image(moving, transform, reference=fixed)

    if fixed.GetDimension() == 3:
        fixed = sitk.MaximumProjection(fixed, 2)[:, :, 0]
        resampled = sitk.MaximumProjection(resampled, 2)[:, :, 0]

    fixed = sitk.Cast(fixed, sitk.sitkUInt8)
    resampled = sitk.Cast(resampled, sitk.sitkUInt8)
    fixed = sitk.RescaleIntensity(fixed, 0, 255)
    resampled = sitk.RescaleIntensity(resampled, 0, 255)

    plt.subplot(1, 2, 2)
    plt.title("Overlay")
    show.display_2d_image_overlay(fixed, resampled)


    del metric_values


def plot_values(registration_method):
    global metric_values

    metric_values.append(registration_method.GetMetricValue())
    # clear the output area (wait=True, to reduce flickering), and plot current data
    # plot the similarity metric values


# endregion

# region RIGID SPATIAL DOMAIN REGISTRATION METHODS

#todo: Make a single method for n-dimensions. Too complicated now

def itk_registration_rigid_3d(fixed_image, moving_image, options):
    """
    A Python implementation for a Rigid Body 3D registration, utilizing
    ITK (www.itk.org) library functions.

    :param fixed_image:     The reference image. Must be an instance of
                            sitk.Image class.
    :param moving_image:    The image for which the spatial transform will
                            be calculated. Same requirements as above.
    :param options:         Options provided by the user via CLI or the
                            included GUI. See image_quality_options.py.
    :return:
                            The final transform as a sitk.Euler2DTransform
    """
    print('Setting up registration job')

    assert isinstance(fixed_image, sitk.Image)
    assert isinstance(moving_image, sitk.Image)

    fixed_image = sitk.Cast(fixed_image, sitk.sitkFloat32)
    moving_image = sitk.Cast(moving_image, sitk.sitkFloat32)

    # REGISTRATION COMPONENTS SETUP
    # ========================================================================

    registration = sitk.ImageRegistrationMethod()

    # OPTIMIZER
    registration.SetOptimizerAsRegularStepGradientDescent(
        options.learning_rate,
        options.min_step_length,
        options.registration_max_iterations,
        relaxationFactor=options.relaxation_factor,
        estimateLearningRate=registration.EachIteration
    )

    registration.SetOptimizerScalesFromJacobian()

    # translation_scale = 1.0/options.translation_scale
    # registration.SetOptimizerScales([1.0, translation_scale, translation_scale])

    # INTERPOLATOR
    registration.SetInterpolator(sitk.sitkLinear)

    # METRIC
    if options.registration_method == 'mattes':
        registration.SetMetricAsMattesMutualInformation(
            numberOfHistogramBins=options.mattes_histogram_bins
        )
    elif options.registration_method == 'correlation':
        registration.SetMetricAsCorrelation()

    elif options.registration_method == 'mean-squared-difference':
        registration.SetMetricAsMeanSquares()
    else:
        raise ValueError("Unknown metric: %s" % options.registration_method)

    registration.SetMetricSamplingStrategy(registration.RANDOM)
    registration.SetMetricSamplingPercentage(options.sampling_percentage)

    if options.reg_translate_only:
        tx = sitk.TranslationTransform(3)
    else:

        tx = sitk.Euler3DTransform()

        transform = sitk.CenteredTransformInitializer(
            fixed_image,
            moving_image,
            tx,
            sitk.CenteredTransformInitializerFilter.MOMENTS
        )
        registration.SetInitialTransform(transform)

        tx.SetCenter(ops_itk.calculate_center_of_image(moving_image))

    registration.SetInitialTransform(tx)

    if options.reg_enable_observers:
        # OBSERVERS
        registration.AddCommand(sitk.sitkStartEvent, start_plot)
        registration.AddCommand(sitk.sitkIterationEvent, lambda: plot_values(registration))

    # START
    # ========================================================================

    print("Starting registration")
    final_transform = registration.Execute(fixed_image, moving_image)

    print(('Final metric value: {0}'.format(registration.GetMetricValue())))
    print(('Optimizer\'s stopping condition, {0}'.format(registration.GetOptimizerStopConditionDescription())))

    if options.reg_enable_observers:
        end_plot(fixed_image, moving_image, final_transform)

    return final_transform


def itk_registration_rigid_2d(fixed_image, moving_image, options):
    """
    A Python implementation for a Rigid Body 2D registration, utilizing
    ITK (www.itk.org) library functions.

    :param fixed_image:     The reference image. Must be an instance of
                            sitk.Image class.
    :param moving_image:    The image for which the spatial transform will
                            be calculated. Same requirements as above.
    :param options:         Options provided by the user via CLI or the
                            included GUI. See image_quality_options.py.
    :return:
                            The final transform as a sitk.Euler2DTransform
    """
    if options.verbose:
        print('Setting up registration job')

    assert isinstance(fixed_image, sitk.Image)
    assert isinstance(moving_image, sitk.Image)

    fixed_image = sitk.Cast(fixed_image, sitk.sitkFloat32)
    moving_image = sitk.Cast(moving_image, sitk.sitkFloat32)

    # REGISTRATION COMPONENTS SETUP
    # ========================================================================

    registration = sitk.ImageRegistrationMethod()

    # OPTIMIZER
    registration.SetOptimizerAsRegularStepGradientDescent(
        options.learning_rate,
        options.min_step_length,
        options.registration_max_iterations,
        relaxationFactor=options.relaxation_factor,
        estimateLearningRate=registration.EachIteration
    )

    translation_scale = 1.0/options.translation_scale
    registration.SetOptimizerScales([1.0, translation_scale, translation_scale])

    # INTERPOLATOR
    registration.SetInterpolator(sitk.sitkLinear)

    # METRIC
    if options.registration_method == 'mattes':
        registration.SetMetricAsMattesMutualInformation(
            numberOfHistogramBins=options.mattes_histogram_bins
        )
    elif options.registration_method == 'correlation':
        registration.SetMetricAsCorrelation()

    elif options.registration_method == 'mean-squared-difference':
        registration.SetMetricAsMeanSquares()
    else:
        raise ValueError("Unknown metric: %s" % options.registration_method)

    registration.SetMetricSamplingStrategy(registration.RANDOM)
    registration.SetMetricSamplingPercentage(options.sampling_percentage)

    if options.reg_translate_only:
        tx = sitk.TranslationTransform(2)
    else:

        tx = sitk.Euler2DTransform()
        tx.SetAngle(options.set_rotation)
        if options.initializer:
            if options.verbose:
                print('Calculating initial registration parameters')
            transform = sitk.CenteredTransformInitializer(
                fixed_image,
                moving_image,
                tx,
                sitk.CenteredTransformInitializerFilter.GEOMETRY
            )
            registration.SetInitialTransform(transform)

        else:
            tx.SetTranslation([options.y_offset, options.x_offset])

            tx.SetCenter(ops_itk.calculate_center_of_image(moving_image))
    registration.SetInitialTransform(tx)

    if options.reg_enable_observers:
        # OBSERVERS
        registration.AddCommand(sitk.sitkStartEvent, start_plot)
        registration.AddCommand(sitk.sitkIterationEvent, lambda: plot_values(registration))

    # START
    # ========================================================================
    if options.verbose:
        print("Starting registration")
    final_transform = registration.Execute(fixed_image, moving_image)

    if options.verbose:
        print(('Final metric value: {0}'.format(registration.GetMetricValue())))
        print(('Optimizer\'s stopping condition, {0}'.format(registration.GetOptimizerStopConditionDescription())))

    if options.reg_enable_observers:
        end_plot(fixed_image, moving_image, final_transform)

    return final_transform

# endregion

# region DEFORMABLE SPATILA DOMAIN REGISTRATION METHDOS

def itk_registration_similarity_2d(fixed_image, moving_image, options):
    """
    A Python implementation for a Rigid Body 2D registration, utilizing
    ITK (www.itk.org) library functions.

    :param fixed_image:     The reference image. Must be an instance of
                            sitk.Image class.
    :param moving_image:    The image that is to be registered. Must be
                            an instance of sitk.Image class.
                            The image for which the spatial transform will
                            be calculated. Same requirements as above.
    :param options:         Options provided by the user via CLI

    :return:                The final transform as a sitk.Similarity2DTransform
    """
    print('Setting up registration job')

    assert isinstance(fixed_image, sitk.Image)
    assert isinstance(moving_image, sitk.Image)

    fixed_image = sitk.Cast(fixed_image, sitk.sitkFloat32)
    moving_image = sitk.Cast(moving_image, sitk.sitkFloat32)

    # REGISTRATION COMPONENTS SETUP
    # ========================================================================

    registration = sitk.ImageRegistrationMethod()

    # OPTIMIZER
    registration.SetOptimizerAsRegularStepGradientDescent(
        options.max_step_length,
        options.min_step_length,
        options.registration_max_iterations,
        relaxationFactor=options.relaxation_factor
    )
    translation_scale = 1.0 / options.translation_scale
    scaling_scale = 1.0 / options.scaling_scale

    registration.SetOptimizerScales([scaling_scale, 1.0, translation_scale, translation_scale])

    # INTERPOLATOR
    registration.SetInterpolator(sitk.sitkLinear)

    # METRIC
    if options.registration_method == 'mattes':
        registration.SetMetricAsMattesMutualInformation(
            numberOfHistogramBins=options.mattes_histogram_bins
        )
        registration.SetMetricSamplingStrategy(registration.RANDOM)
        registration.SetMetricSamplingPercentage(options.mattes_sampling_percentage)

    elif options.registration_method == 'correlation':
        registration.SetMetricAsCorrelation()

    elif options.registration_method == 'mean-squared-difference':
        registration.SetMetricAsMeanSquares()
    else:
        raise ValueError("Unknown metric: %s" % options.registration_method)

    print('Calculating initial registration parameters')
    tx = sitk.Similarity2DTransform()
    tx.SetAngle(options.set_rotation)
    tx.SetScale(options.set_scale)

    if options.initializer:
        transform = sitk.CenteredTransformInitializer(
            fixed_image,
            moving_image,
            tx,
            sitk.CenteredTransformInitializerFilter.GEOMETRY
        )
        registration.SetInitialTransform(transform)
    else:
        tx.SetTranslation([options.y_offset, options.x_offset])
        tx.SetCenter(ops_itk.calculate_center_of_image(moving_image))
        registration.SetInitialTransform(tx)

    # OBSERVERS

    registration.AddCommand(sitk.sitkStartEvent, start_plot)
    #registration.AddCommand(sitk.sitkEndEvent, end_plot)
    registration.AddCommand(sitk.sitkIterationEvent, lambda: plot_values(registration))

    # START
    # ========================================================================

    print("Starting registration")
    final_transform = registration.Execute(fixed_image, moving_image)

    print(('Final metric value: {0}'.format(registration.GetMetricValue())))
    print(('Optimizer\'s stopping condition, {0}'.format(registration.GetOptimizerStopConditionDescription())))

    end_plot(fixed_image, moving_image, final_transform)

    return final_transform


def itk_registration_affine_2d(fixed_image, moving_image, options):
    """
    A Python implementation for a Rigid Body 2D registration, utilizing
    ITK (www.itk.org) library functions.

    :param fixed_image:     The reference image. Must be an instance of
                            sitk.Image class.
    :param moving_image:    The image that is to be registered. Must be
                            an instance of sitk.Image class.
                            The image for which the spatial transform will
                            be calculated. Same requirements as above.
    :param options:         Options provided by the user via CLI

    :return:                The final transform as a sitk.Similarity2DTransform
    """
    print('Setting up registration job')

    assert isinstance(fixed_image, sitk.Image)
    assert isinstance(moving_image, sitk.Image)

    fixed_image = sitk.Cast(fixed_image, sitk.sitkFloat32)
    moving_image = sitk.Cast(moving_image, sitk.sitkFloat32)

    # REGISTRATION COMPONENTS SETUP
    # ========================================================================

    registration = sitk.ImageRegistrationMethod()

    # OPTIMIZER
    registration.SetOptimizerAsRegularStepGradientDescent(
        options.max_step_length,
        options.min_step_length,
        options.registration_max_iterations,
        relaxationFactor=options.relaxation_factor
    )
    translation_scale = 1.0 / options.translation_scale
    scaling_scale = 1.0 / options.scaling_scale

    registration.SetOptimizerScales([scaling_scale, 1.0, translation_scale, translation_scale])

    # INTERPOLATOR
    registration.SetInterpolator(sitk.sitkLinear)

    # METRIC
    if options.registration_method == 'mattes':
        registration.SetMetricAsMattesMutualInformation(
            numberOfHistogramBins=options.mattes_histogram_bins
        )
        registration.SetMetricSamplingStrategy(registration.RANDOM)
        registration.SetMetricSamplingPercentage(options.mattes_sampling_percentage)

    elif options.registration_method == 'correlation':
        registration.SetMetricAsCorrelation()

    elif options.registration_method == 'mean-squared-difference':
        registration.SetMetricAsMeanSquares()
    else:
        raise ValueError("Unknown metric: %s" % options.registration_method)

    print('Calculating initial registration parameters')
    tx = sitk.AffineTransform()


    if options.initializer:
        transform = sitk.CenteredTransformInitializer(
            fixed_image,
            moving_image,
            tx,
            sitk.CenteredTransformInitializerFilter.MOMENTS
        )
        registration.SetInitialTransform(transform)
    else:
        tx.SetTranslation([options.y_offset, options.x_offset])
        tx.SetCenter(ops_itk.calculate_center_of_image(moving_image))
        registration.SetInitialTransform(tx)

    # OBSERVERS

    registration.AddCommand(sitk.sitkStartEvent, start_plot)
    #registration.AddCommand(sitk.sitkEndEvent, end_plot)
    registration.AddCommand(sitk.sitkIterationEvent, lambda: plot_values(registration))

    # START
    # ========================================================================

    print("Starting registration")
    final_transform = registration.Execute(fixed_image, moving_image)

    print(('Final metric value: {0}'.format(registration.GetMetricValue())))
    print(('Optimizer\'s stopping condition, {0}'.format(registration.GetOptimizerStopConditionDescription())))

    end_plot(fixed_image, moving_image, final_transform)

    return final_transform

# endregion

# region RIGID FREQUENCY DOMAIN REGISTRATION METHODS


def phase_correlation_registration(fixed_image, moving_image, subpixel=100, verbose=False, resample=True):
    """
    A simple Phase Correlation based image registration method.
    :param verbose:  enable print functions
    :param subpixel: resampling factor; registration will be perfromed to
                     1/subpixel accuracy
    :param fixed_image: the reference image as MIPLIB Image object
    :param moving_image: the moving image as MIPLIB Image object
    :return: returns the SimpleITK transform
    """

    assert isinstance(fixed_image, Image)
    assert isinstance(moving_image, Image)

    shift, error, diffphase = register_translation(fixed_image, moving_image, subpixel)

    scaled_shifts = list(-offset * spacing for offset, spacing in zip(shift, fixed_image.spacing))
    if verbose:
        print(("Detected offset (y, x): {}".format(scaled_shifts)))

    resampled = np.abs(np.fft.ifftn(fourier_shift(np.fft.fftn(moving_image),
                                                  shift)).real)
    if resample:
        return Image(resampled, fixed_image.spacing)
    else:
        return scaled_shifts

# endregion



#!/usr/bin/env python
# -*- python -*-

"""
File: subjective.py
Author: Sami Koho (sami.koho@gmail.com)

Description:

A utility script for performing subjective image rankings. One image is
displayed at a time, after which it is ranked on 1-5 scale, 5 being the
best and 1 the worst. The script can be run multiple times to collect
several ranking results in a single csv file. At every run the data
is shuffled in order to not repeat the same image sequence twice.
"""

import os
import sys

import matplotlib.pyplot as plt
import pandas

import miplib.ui.cli.miplib_entry_point_options as script_options


def main():

    options = script_options.get_subjective_ranking_options(sys.argv[1:])
    path = options.working_directory
    index = 0
    assert os.path.isdir(path), path

    # Create or open a csv file
    output_dir = path
    file_name = "subjective_ranking_scores.csv"
    file_path = os.path.join(output_dir, file_name)

    if os.path.exists(file_path):
        csv_data = pandas.read_csv(file_path)
        # Append a new result column
        for column in csv_data:
            if "Result" in column:
                index += 1
    else:
        csv_data = pandas.DataFrame()
        file_names = []
        # Get valid file names
        for image_name in os.listdir(path):
            real_path = os.path.join(path, image_name)
            if not os.path.isfile(real_path) or not real_path.endswith((".jpg", ".tif", ".tiff", ".png")):
                continue
            file_names.append(image_name)
        csv_data["Filename"] = file_names

    result_name = "Result_" + str(index)
    results = []

    # Plot settings
    plt.ion()
    plt.axis('off')

    # Shuffle the data frame so that the order of the displayed images is mixed every time.
    csv_data = csv_data.sample(frac=1)
    print("Images are graded on a scale 1-5, where 1 denotes a very bad image " \
          "and 5 an excellent image")

    for image_name in csv_data["Filename"]:
        real_path = os.path.join(path, image_name)
        image = plt.imread(real_path)

        plt.imshow(image, cmap='hot', vmax=image.max(), vmin=image.min())

        success = False
        while not success:
            input = input("Give grade: ")

            if input.isdigit():
                result = int(input)
            else:
                print("Please give a numeric grade 1-5.")
                continue

            if 1 <= result <= 5:
                success = True
                results.append(result)
            else:
                print("Please give a numeric grade 1-5.")

    csv_data[result_name] = results
    csv_data.to_csv(file_path, index=False)


if __name__ == "__main__":
    main()














import os

import h5py

from miplib.data.containers.fourier_correlation_data import FourierCorrelationDataCollection, FourierCorrelationData
from miplib.data.containers.image import Image


class FourierCorrelationDataReader(object):
    """
    A class for writing Fourier Correlation Data into a file.
    """

    # region Constructor and Destructor
    def __init__(self, file_path):

        # Create output dir, if it doesn't exist output dir if
        if not os.path.isfile(file_path) or not file_path.endswith(".hdf5"):
            raise ValueError("Not a valid filename: %s" % file_path)

        self.data = h5py.File(file_path, mode="r")

    def __del__(self):
        self.close()

    # endregion

    def read_metadata(self):
        """
        Read a metadata dictionary from the HDF5 files header
        """
        return dict(self.data.attrs)

    def read_images(self, index=None):
        """
        Read images from the data structure.
        :returns images: a tuple of Image objects, or a single image
        """

        if "images" not in self.data:
            raise ValueError("No images to read")

        if index is not None:
            image_name = "image_%i" % index
            data_set = self.data["images"][image_name]
            spacing = data_set.attrs["pixel_size"].split()[0:len(data_set.shape)]
            return Image(data_set[:], spacing)

        images = []
        for data_set in self.data["images"]:
            spacing = data_set.attrs["pixel_size"].split()[0:len(data_set.shape)]
            images.append(Image(data_set[:], spacing))

        return images

    def read_data_set(self):
        """
        Read Fourier Correlation Data file (FRC, FSC etc) to the FourierCorrelationDataCollection
        data structure.
        :returns FourierCorrelationDataCollection
        """
        group_prefix = "data_set_"
        data_sets = FourierCorrelationDataCollection()
        for group_name in list(self.data.keys()):
            if group_prefix in group_name:
                angle = group_name.split("_")[-1]
                data_set = FourierCorrelationData()
                resolution_group = self.data[group_name]["resolution"]
                correlation_group = self.data[group_name]["correlation"]

                data_set.resolution["threshold"] = resolution_group["threshold"][:]
                data_set.resolution["resolution-point"] = \
                    resolution_group.attrs["resolution-point"].split()[:-1]
                data_set.resolution["criterion"] = resolution_group.attrs["criterion"]
                data_set.resolution["resolution-threshold-coefficients"] = \
                    resolution_group["resolution-threshold-coefficients"][:]

                data_set.correlation["correlation"] = correlation_group["correlation"][:]
                data_set.correlation["frequency"] = correlation_group["frequency"][:]
                data_set.correlation["points-x-bin"] = correlation_group["points-x-bin"][:]
                data_set.correlation["curve-fit"] = correlation_group["curve-fit"][:]
                data_set.correlation["curve-fit-coefficients"] = \
                    correlation_group["curve-fit-coefficients"][:]

                data_sets[int(angle)] = data_set

        return data_sets

    def close(self):
        """
        A function to explicitly close the data file (will be called by the destructor, if not.
        :return:
        """
        self.data.close()




# coding=utf-8
"""
deconvolve.py

Copyright (C) 2016 Sami Koho
All rights reserved.

This software may be modified and distributed under the terms
of the BSD license.  See the LICENSE file for details.

This file contains the GPU accelerated miplib multi-view image fusion
algorithms. The MultiViewFusionRLCuda class implements all the same methods
as the MultiViewFusionRL class, for non-accelerated iterative image fusion.

"""

import itertools

import numpy
import miplib.processing.ops_ext as ops_ext
from numba import cuda, vectorize
from pyculib.fft import FFTPlan, fft_inplace, ifft_inplace

from . import deconvolve
import miplib.processing.ndarray as ops_array


class DeconvolutionRLCuda(deconvolve.DeconvolutionRL):
    """
    This class implements GPU accelerated versions of the iterative image
    fusion algorithms discussed in:

    Koho, S., Deguchi, T., & Hanninen, P. E. (2015). A software tool for
    tomographic axial superresolution in STED microscopy. Journal of Microscopy,
    260(2), 208–218. http://doi.org/10.1111/jmi.12287

    MultiViewFusionRLCuda is inherits most of its functionality from
    MultiViewFusionRL (see deconvolve.py).
    """
    def __init__(self, image, psf, writer, options):
        """
        :param image:   the image as a Image object
        :param psf:     the psf as an Image object

        :param options: command line options that control the behavior
                        of the fusion algorithm
        """
        deconvolve.DeconvolutionRL.__init__(self, image, psf, writer, options)

        padded_block_size = self.block_size + 2*self.options.block_pad
        if self.imdims == 2:
            threadpergpublock = 32, 32
        else:
            threadpergpublock = 32, 32, 8

        blockpergrid = self.__best_grid_size(
            tuple(reversed(padded_block_size)), threadpergpublock)

        FFTPlan(padded_block_size, itype=numpy.complex64, otype=numpy.complex64)


        print(('Optimal kernel config: %s x %s' % (blockpergrid, threadpergpublock)))

        self.__get_fourier_psfs()

    def compute_estimate(self):
        """
            Calculates a single RL fusion estimate. There is no reason to call this
            function -- it is used internally by the class during fusion process.
        """
        print('Beginning the computation of the %i. estimate' % \
              (self.iteration_count + 1))

        self.estimate_new[:] = numpy.zeros(self.image_size, dtype=numpy.float32)

        # Iterate over blocks
        stream1 = cuda.stream()
        stream2 = cuda.stream()

        iterables = (range(0, m, n) for m, n in zip(self.image_size, self.block_size))
        pad = self.options.block_pad
        block_idx = tuple(slice(pad, pad + block) for block in self.block_size)

        if self.imdims == 2:
            for pos in itertools.product(*iterables):

                estimate_idx = tuple(slice(j, j + k) for j, k in zip(idx, self.block_size))
                index = numpy.array(pos, dtype=int)

                if self.options.block_pad > 0:
                    h_estimate_block = self.get_padded_block(self.estimate, index.copy()).astype(numpy.complex64)
                else:
                    h_estimate_block = self.estimate[estimate_idx].astype(numpy.complex64)

                d_estimate_block = cuda.to_device(h_estimate_block, stream=stream1)
                d_psf = cuda.to_device(self.psf_fft, stream=stream2)

                # Execute: cache = convolve(PSF, estimate), non-normalized
                fft_inplace(d_estimate_block, stream=stream1)
                stream2.synchronize()

                self.vmult(d_estimate_block, d_psf, out=d_estimate_block)
                ifft_inplace(d_estimate_block)

                h_estimate_block_new = d_estimate_block.copy_to_host()

                # Execute: cache = data/cache
                h_image_block = self.get_padded_block(self.image, index.copy()).astype(numpy.float32)
                ops_ext.inverse_division_inplace(h_estimate_block_new, h_image_block)

                d_estimate_block = cuda.to_device(h_estimate_block_new,
                                                  stream=stream1)
                d_adj_psf = cuda.to_device(self.adj_psf_fft, stream=stream2)

                fft_inplace(d_estimate_block, stream=stream1)
                stream2.synchronize()
                self.vmult(d_estimate_block, d_adj_psf, out=d_estimate_block)
                ifft_inplace(d_estimate_block)
                h_estimate_block_new = d_estimate_block.copy_to_host().real

                self.estimate_new[estimate_idx] = h_estimate_block_new[block_idx]

        # TV Regularization (doesn't seem to do anything miraculous).
        if self.options.rltv_lambda > 0 and self.iteration_count > 0:
            dv_est = ops_ext.div_unit_grad(self.estimate, self.image_spacing)
            with numpy.errstate(divide="ignore"):
                self.estimate_new /= (1.0 - self.options.rltv_lambda * dv_est)
                self.estimate_new[self.estimate_new == numpy.inf] = 0.0
                self.estimate_new[:] = numpy.nan_to_num(self.estimate_new)

        # Update estimate inplace. Get convergence statistics.
        return ops_ext.update_estimate_poisson(self.estimate,
                                               self.estimate_new,
                                               self.options.convergence_epsilon)

    @staticmethod
    def __best_grid_size(size, tpb):
        bpg = numpy.ceil(numpy.array(size, dtype=numpy.float) / tpb).astype(numpy.int).tolist()
        return tuple(bpg)

    @staticmethod
    @vectorize(['complex64(complex64, complex64)'], target='cuda')
    def vmult(a, b):
        """
        Implements complex array multiplication on GPU

        Parameters
        ----------
        :param  a  Two Numpy arrays of the same shape, dtype=numpy.complex64
        :param  b

        Returns
        -------

        a*b

        """
        return a * b

    @staticmethod
    @vectorize(['float32(float32, float32)'], target='cuda')
    def float_vmult(a, b):
        """
        Implements array multiplication on GPU

        Parameters
        ----------
        :param  a  Two Numpy arrays of the same shape, dtype=numpy.float32
        :param  b

        Returns
        -------

        a*b

        """

        return a * b

    def __get_fourier_psfs(self):
        """
        Pre-calculates the PSFs during image fusion process.
        """
        print("Pre-calculating PSFs")

        psf = self.psf[:]
        if self.imdims == 3:
            adj_psf = psf[::-1, ::-1, ::-1]
        else:
            adj_psf = psf[::-1, ::-1]

        padded_block_size = tuple(self.block_size + 2*self.options.block_pad)

        self.psfs_fft = ops_array.expand_to_shape(psf, padded_block_size).astype(numpy.complex64)
        self.adj_psf_fft = ops_array.expand_to_shape(adj_psf, padded_block_size).astype(numpy.complex64)
        self.psf_fft = numpy.fft.fftshift(self.psf_fft)
        self.adj_psf_fft = numpy.fft.fftshift(self.adj_psf_fft)

        fft_inplace(self.psf_fft)
        fft_inplace(self.adj_psf_fft)









import os

from multiprocessing import Queue

from miplib.data.containers.image import Image
from miplib.data.io import write as imwrite


class ImageWriterBase(object):
    def write(self, image):
        pass


class QueuedImageWriter(ImageWriterBase):
    def __init__(self, queue):
        assert isinstance(queue, Queue)

        self.queue = queue

    def write(self, image):
        assert isinstance(image, Image)

        self.queue.put(image)


class TiffImageWriter(ImageWriterBase):
    def __init__(self, directory):
        self.index = 0
        self.dir = directory

    def __get_full_path(self):
        filename = "result_{}.tif".format(self.index)
        return os.path.join(self.dir, filename)

    def write(self, image):
        assert isinstance(image, Image)

        imwrite.image(self.__get_full_path(), image)

        self.index += 1






import os

import SimpleITK as sitk
import pims

import miplib.processing.itk as itkutils
from . import tiffile
from miplib.data.containers.image import Image

scale_c = 1.0e6


def get_image(filename, series=0, channel=0, return_type='image', bioformats=True):
    """
    A wrapper for the image read functions.
    Parameters
    :param series: The index of an image in a time series
    :param channel: The color channel in  a multi-channel image.
    :param bioformats: Toggle to disable bioformats reader.
    :param filename The full path to an image
    :param return_type Return the image as miplib Image. sitk.Image.
           the return type can be chosen with a string ('image, 'itk').

    """
    assert return_type in ('itk', 'image')

    if filename.endswith(".mha") or not bioformats:
        data = __itk_image(filename, return_type == 'itk')
    else:
        data = __bioformats(filename, series, channel, return_type == 'itk')

    return data

def __itk_image(filename, return_itk=True):
    """
    A function for reading image file types typical to ITK (mha & mhd). This is mostly
    of a historical significance, because in the original miplib 1 such files were
    used, mostly for convenience.

    :param filename:     Path to an ITK image
    :param return_itk    Toggle whether to convert the ITK image into Numpy format
    :return:             Image data as a Numpy array, voxel spacing tuple
    """
    assert filename.endswith((".mha", ".mhd"))
    image = sitk.ReadImage(filename)
    if return_itk:
        return image
    else:
        return itkutils.convert_from_itk_image(image)


def __tiff(filename, memmap=False, return_itk=False):
    """
    ImageJ has a bit peculiar way of saving image metadata, especially the tags
    for voxel spacing, which is of main interest in miplib. This function reads
    a 3D TIFF into a Numpy array and also calculates the voxel spacing parameters
    from the TIFF tags. I will not guarantee that this will work with any other TIFF
    files.

    :param filename:    Path to a TIFF.
    :param memmap:      Enables Memory mapping in case the TIFF file is too large to
                        be read in memory completely.
    :param return_itk  Converts the Image data into a sitk.Image. This can be used
                        when single images are needed, instead of using the HDF5
                        structure adopted in miplib.
    :return:            Image data either as a Numpy array, voxel spacing tuple or a
                        sitk.Image
    """
    assert filename.endswith((".tif", ".tiff"))
    tags = {}
    # Read images and tags
    with tiffile.TiffFile(filename) as image:
        # Get images
        images = image.asarray(memmap=memmap)
        # Get tags
        page = image[0]
        for tag in list(page.tags.values()):
            tags[tag.name] = tag.value

    # Figure out z-spacing, which in ImageJ is hidden in the "image_description"
    # header (why, one might ask).
    image_descriptor = tags["image_description"].split("\n")
    z_spacing = None
    for line in image_descriptor:
        if "spacing" in line:
            z_spacing = float(line.split("=")[-1])
            break
    assert z_spacing is not None

    # Create a tuple for zxy-spacing. The order of the dimensions follows that of the
    # image data
    spacing = (z_spacing, scale_c/tags["x_resolution"][0], scale_c/tags[
        "y_resolution"][0])

    #print spacing
    if return_itk:
        return itkutils.convert_from_numpy(images, spacing)
    else:
        return images, spacing


def __itk_transform(path, return_itk=False):
    """
    Prior to starting to use the HDF5 format data storage images and spatial
    transforms were saved as separate image files on the hard drive. This
    function can be used to read a spatial transform saved from ITK. It is
    to transfer old files into the HDF5 format storage.

    Parameters
    ----------
    path        Path to the transform file (usually txt ended)

    Returns     Returns the transform type integer, parameters and fixed
                parameters.
    -------

    """

    if not os.path.isfile(path):
        raise ValueError("Not a valid path: %s" % path)

    transform = sitk.ReadTransform(path)

    if return_itk:
        return transform

    else:
        # #TODO: Check that this makes any sense. Also consult the ITK HDF implementation for ideas
        # with open(path, 'r') as f:
        #     for line in f:
        #         if line.startswith('Transform:'):
        #             type_string = line.split(': ')[1].split('_')[0]
        #             if "VersorRigid" in type_string:
        #                 transform_type = itk_transforms_c['sitkVersorRigid']
        #                 break
        #             else:
        #                 raise NotImplementedError("Unknown transform type: "
        #                                           "%s" % type_string)
        transform_type = transform.GetName()
        params = transform.GetParameters()
        fixed_params = transform.GetFixedParameters()
        return transform_type, params, fixed_params


def __bioformats(filename, series=0, channel=0, return_itk = False):
    """
    Read an image using the Bioformats importer. Good for most microscopy formats.

    :param filename:
    :param series:
    :param return_itk:
    :return:
    """
    assert pims.bioformats.available(), "Please install jpype in order to use " \
                                        "the bioformats reader."
    image = pims.bioformats.BioformatsReader(filename, series=series)

    # Get Pixel/Voxel size information
    if 'z' not in image.axes:
        spacing = (image.metadata.PixelsPhysicalSizeY(0),
                   image.metadata.PixelsPhysicalSizeX(0))
    else:
        spacing = (image.metadata.PixelsPhysicalSizeZ(0),
                   image.metadata.PixelsPhysicalSizeY(0),
                   image.metadata.PixelsPhysicalSizeX(0))

    # Get color channel
    if 'c' in image.sizes:
        image.iter_axes = 'c'
        assert len(image) > channel
        image = image[channel]
    else:
        image = image[0]

    if return_itk:
        return itkutils.convert_from_numpy(image, spacing)
    else:
        return Image(image, spacing)






from matplotlib import pyplot as plt

def save_figure(figure, path, dpi=1200):
    """
    A really simple utility to save a figure to file.
    :param figure: a matplotlib.pyplot.Figure instance
    :param path:   a full path to the file. Make sure that the directory exists. The file type
                   will be decided according to the filename suffix.
    :param dpi:    dpi value for the plot
    :return:       nothing
    """
    assert isinstance(figure, plt.Figure)

    figure.savefig(path, dpi=dpi, bbox_inches='tight', pad_inches=0, transparent=True)



from miplib.data.containers.fourier_correlation_data import FourierCorrelationDataCollection
import miplib.data.iterators.fourier_ring_iterators as iterators
import miplib.processing.image as imops
import miplib.analysis.resolution.analysis as frc_analysis
import miplib.analysis.resolution.fourier_ring_correlation as frc

from .psfgen import PsfFromFwhm


def generate_frc_based_psf(image, args):
    fwhm = [frc.calculate_single_image_frc(image, args).resolution["resolution"], ] * 2
    psf_generator = PsfFromFwhm(fwhm)

    if image.ndim == 2:
        return psf_generator.xy()
    else:
        return psf_generator.volume()




import os
import sys
import time

from pyculib import cuda_compatible
import miplib.processing.to_string as ops_output
import miplib.ui.utils as uiutils
from miplib.processing.deconvolution import deconvolve_cuda, deconvolve
from miplib.psf import psfgen
from miplib.ui.cli import miplib_entry_point_options as options
from miplib.data.io import read as imread


def main():
    args = options.get_deconvolve_script_options(sys.argv[1:])

    # Load image
    image_name = args.image
    dir = args.working_directory
    image_path = os.path.join(dir, image_name)

    # Figure out an image type and use the correct loader.
    if image_path.endswith('.tif'):
        image = imread.get_image(image_path)
    elif image_path.endswith('.mat'):
        image = imread.get_image(image_path)[args.carma_gate_idx, args.carma_det_idx]
    else:
        raise AttributeError("Unknown image type %s" % image_path.split('.')[-1])


    # Load/Generate PSF
    if args.psf == "estimate":
        psf = psfgen.PSF(psftype=args.psf_type, shape=args.psf_shape, dims=args.psf_size,
                         ex_wavelen=args.ex_wl, em_wavelen=args.em_wl, num_aperture=args.na,
                         refr_index=args.refractive_index, magnification=args.magnification,
                         pinhole_radius=args.pinhole_radius)

        if args.sted_psf:
            psf.sted_correction(args.sted_phi, args.sted_sigma)

            psf_image = psf.volume()
            psf_spacing = psf.dims['um'][0]/psf.shape[0]

            if image.get_dimensions()[0] == 1:
                psf_image = psf_image[psf_image.shape/2]

            psf = image.Image(psf_image, psf_spacing)

    else:
        psf = image.Image.get_image_from_imagej_tiff(args.psf_path)

    # Start deconvolution
    if cuda_compatible():
        print("Found a compatible GPU. The image deconvolution will be run with " \
              "hardware acceleration.")
        task = deconvolve_cuda.DeconvolutionRLCuda(image, psf, args)
    else:
        task = deconvolve.DeconvolutionRL(image, psf, args)

    begin = time.time()
    task.execute()
    end = time.time()
    result = task.get_result()

    print("Deconvolution complete.")
    print("The deconvolution process with %i iterations " \
          "took %s (H:M:S) to complete." % (args.max_nof_iterations,
                                            ops_output.format_time_string(
                                                end - begin)))
    if uiutils.get_user_input("Do you want to save the result to TIFF? "):
        file_path = os.path.join(args.working_directory,
                                 "fusion_result.tif")
        result.save_to_tiff(file_path)


if __name__ == "__main__":
        main()




#!/usr/bin/env python
# -*- python -*-

"""
File: utils_image_subtractor.py
Author: Sami Koho (sami.koho@gmail.com)
Description:
I happened to receive a dataset, of 300+ HCS images in jpg to be analyzed.
The images contained phase contrast data (Red) and fluorescence data (Green)
The color channels in the jpg files were mixed so that the green channel
contained both red and green, whereas red contained only green. Therefore
in order to get the fluorescence only images, I needed to get rid of the
crosstalk. This is a small utility for doing that; it can
of course quite easily be modified into any kind of a batch processing task.
"""

import os
import sys

from miplib.data.io import read, write


def main():
    path = sys.argv[1]
    assert os.path.isdir(path), path

    # Create output directory
    output_dir = os.path.join(path, "Subtracted")
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)

    for image_name in os.listdir(path):
        real_path = os.path.join(path, image_name)
        if not os.path.isfile(real_path) or not real_path.endswith(".jpg"):
            continue
        image_red = read.get_image(real_path, channel=0)
        image_green = read.get_image(real_path, channel=1)

        image_sub = image_green - image_red

        save_name = "sub_" + image_name
        save_path = os.path.join(path, output_dir)
        save_path = os.path.join(save_path, save_name)
        write.image(image_sub, save_path)
        print("Saved %s to %s" % (save_name, save_path))

if __name__ == "__main__":
    main()


import os
#import subprocess

import SimpleITK as sitk
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
import numpy as np

# import miplib.data.io.tiffile as tiffile
#
# vaa3d_bin = "/home/sami/bin/Vaa3D_Ubuntu_64bit_v3.200/vaa3d"
#
# def evaluate_3d_image(image):
#     """
#     A utility function that can be used to display the registration
#     and/or fusion results in Vaa3D volume viewer. The function returns
#     a Boolean value based on whether the user wants to save the image
#     into the data storage or not.
#
#     Parameters
#     ----------
#     image       A 3D image volume as a numpy.ndarray. The order of the
#                 dimensions should be ZXYC. C can be omitted if one.
#
#     """
#     assert os.path.exists(vaa3d_bin)
#
#     filename = "temp.tif"
#     tiffile.imsave(filename, image)
#
#     subprocess.call([vaa3d_bin, "-i", filename])
#
#     os.remove(filename)


# callback invoked by the interact ipython method for scrolling through the image stacks of
# the two images (moving and fixed)
def display_3d_slices(fixed_image_z, moving_image_z, fixed_npa, moving_npa):
    # create a figure with two subplots and the specified size
    plt.subplots(1, 2, figsize=(10, 8))

    # draw the fixed image in the first subplot
    plt.subplot(1, 2, 1)
    plt.imshow(fixed_npa[fixed_image_z, :, :], cmap='gray')
    plt.title('fixed image')
    plt.axis('off')

    # draw the moving image in the second subplot
    plt.subplot(1, 2, 2)
    plt.imshow(moving_npa[moving_image_z, :, :], cmap='gray')
    plt.title('moving image')
    plt.axis('off')

# callback invoked by the ipython interact method for scrolling and modifying the alpha blending
# of an image stack of two images that occupy the same physical space.


def display_3d_slice_with_alpha(image_z, alpha, fixed, moving):
    img = (1.0 - alpha) * fixed[:, :, image_z] + alpha * moving[:, :, image_z]
    plt.imshow(sitk.GetArrayFromImage(img), cmap='gray')
    plt.axis('off')


def create_axial_views_plot(image, x_idx, y_idx, z_idx):

    assert issubclass(image.__class__, np.ndarray)
    assert image.ndim == 3

    xy = image[z_idx, :, :]
    xz = image[:, y_idx, :]
    yz = image[:, :, x_idx]

    yz = np.transpose(yz)

    width_ratio = xy.shape[1]/yz.shape[1]
    height_ratio = xy.shape[0]/xz.shape[0]

    fig = plt.figure(figsize=(8, 8))
    gs = gridspec.GridSpec(2, 2,
                           width_ratios=[width_ratio, 1],
                           height_ratios=[height_ratio, 1])

    ax0 = plt.subplot(gs[0, 0])
    ax0.imshow(xy, cmap="hot")
    ax0.set_title("XY")
    ax0.axis('off')

    ax1 = plt.subplot(gs[0, 1])
    ax1.imshow(yz, cmap="hot")
    ax1.set_title("YZ")
    ax1.axis('off')

    ax2 = plt.subplot(gs[1,0])
    ax2.imshow(xz, cmap="hot")
    ax2.set_title("XZ")
    ax2.axis('off')

    #fig.delaxes(axes[1, 1])
    return fig



def display_2d_images(image1,
                      image2,
                      image1_title='image1',
                      image2_title='image2',
                      vertical=False):
    """
    A function that can be used to display two SimpleITK images side by side.
    It is also possible to select paired landmarks from the two images, by
    enabling the landmarks argument.

    Parameters

    image1      A numpy.ndarray or its subclass
    image2      A numpy.ndarray or its subclass

    """
    assert issubclass(type(image1), np.ndarray)
    assert issubclass(type(image2), np.ndarray)

    assert image1.ndim == 2 and image2.ndim == 2

    if vertical:
        fig, (ax1, ax2) = plt.subplots(
            2, 1, figsize=(13, 10),
            gridspec_kw = {'height_ratios':[3, 1], 'width_ratios': [1, 1]}
        )
    else:
        fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(14, 10))

    # draw the fixed image in the first subplot
    ax1.imshow(image1, cmap="hot")
    ax1.set_title(image1_title)
    ax1.axis('off')

    # draw the moving image in the second subplot
    ax2.imshow(image2, cmap="hot")
    ax2.set_title(image2_title)
    ax2.axis('off')

    plt.show()


def display_2d_image(image):
    """
    A function that can be used to display two SimpleITK images side by side.
    It is also possible to select paired landmarks from the two images, by
    enabling the landmarks argument.

    Parameters

    image       a Numpy array or a SimpleITk image object

    """

    if isinstance(image, sitk.Image):
        image = sitk.GetArrayFromImage(image)

    assert image.ndim == 2

    plt.imshow(image, cmap="rainbow")
    plt.axis('off')
    plt.show()


def display_2d_slices_with_alpha(alpha, fixed, moving):
    img = (1.0 - alpha) * fixed + alpha * moving
    plt.imshow(sitk.GetArrayFromImage(img), cmap='gray')
    plt.axis('off')


def display_2d_image_overlay(image1, image2, image3=None):
    '''
    Overlays 2-3 images into a single RGB plot. This was intended for use in
    evaluating registration results.
    Parameters
    ----------
    image1      A 2D numpy.array or sitk.Image that
    image2      A 2D numpy.array or sitk.Image that
    image3      A 2D numpy.array or sitk.Image that

    Returns     Nothing
    -------

    '''
    if isinstance(image1, sitk.Image):
        image1 = sitk.GetArrayFromImage(image1)
    if isinstance(image2, sitk.Image):
        image2 = sitk.GetArrayFromImage(image2)

    if image1.shape != image2.shape:
        raise ValueError("The dimensions of the images to be overlaid should match")

    if image3 is None:
        image3 = np.zeros(image1.shape, dtype=np.uint8)

    rgb_image = np.concatenate([aux[..., np.newaxis] for aux in (image1, image2, image3)], axis=-1)

    plt.imshow(rgb_image)
    plt.axis('off')
    plt.show()


def show_pics_from_disk(filenames, title="Image collage"):
    """
    A utility for creating a collage of images, to be shown
    in a single plot. The images are loaded from disk according
    to the provided filenames:
    :param filenames:   A list containing the image filenames
    :param title:       Name of the plot
    :return:            Nothing
    """
    if len(filenames) > 1:
        if 4 < len(filenames) <= 9:
            fig, subplots = plt.subplots(3, 3)
        elif 9 < len(filenames) <= 16:
            fig, subplots = plt.subplots(4, 4)
        elif 16 < len(filenames) <= 25:
            fig, subplots = plt.subplots(5, 5)
        elif 25 < len(filenames) <= 36:
            fig, subplots = plt.subplots(6, 6)
        else:
            fig, subplots = plt.subplots(2, 2)

        # fig.title(title)
        i = 0
        j = 0
        k = 0
        while k < len(filenames):
            j = 0
            while j < subplots.shape[1] and k < len(filenames):
                print(filenames[i + j])
                subplots[i, j].imshow(plt.imread(filenames[k]), cmap='hot')
                subplots[i, j].set_title(os.path.basename(filenames[k]))
                subplots[i, j].axis("off")
                k += 1
                j += 1
            i += 1
        plt.subplots_adjust(wspace=-0.5, hspace=0.2)
        plt.suptitle(title, size=16)
        plt.show()

    else:
        plt.imshow(plt.imread(filenames))
        plt.axis("off")
        plt.show()


import SimpleITK as sitk
import numpy

import miplib.processing.itk as ops_itk
from miplib.data.containers import image_data
from . import registration


# todo: This class has way too many responsibilities. Need to refactor at some point.
class RotatedMultiViewRegistration(object):
    """
    A class for multiview image registration. The method is based on
    functions inside the Insight Toolkit (www.itk.org), as in the original
    *miplib*. In *miplib* SimpleITK was used instead of Python
    wrapped ITK.

    The registration was updated to support multiple views
    and the new HDF5 data storage implementation. It was also implemented
    as a class.
    """

    def __init__(self, data, options):
        """
        :param data:    a ImageData object

        :param options: command line options that control the behavior
                            of the registration algorithm
         """
        assert isinstance(data, image_data.ImageData)

        # Parameters
        self.data = data
        self.options = options

        # Fixed and moving image
        self.fixed_index = 0
        self.moving_index = 1

        # Results
        self.final_transform = None

        # REGISTRATION COMPONENTS SETUP
        # ========================================================================

        self.registration = sitk.ImageRegistrationMethod()

        # OPTIMIZER

        self.registration.SetOptimizerAsRegularStepGradientDescent(
            options.learning_rate,
            options.min_step_length,
            options.registration_max_iterations,
            relaxationFactor=options.relaxation_factor,
            estimateLearningRate=self.registration.EachIteration
        )
        # translation_scale = 1.0 / options.translation_scale
        # self.registration.SetOptimizerScales([10, 10, 10,
        #                                       .1, .1,.1])

        self.registration.SetOptimizerScalesFromJacobian()

        # INTERPOLATOR
        self.registration.SetInterpolator(sitk.sitkLinear)

        # METRIC
        if options.registration_method == 'mattes':
            self.registration.SetMetricAsMattesMutualInformation(
                numberOfHistogramBins=options.mattes_histogram_bins
            )

        elif options.registration_method == 'correlation':
            self.registration.SetMetricAsCorrelation()

        elif options.registration_method == 'mean-squared-difference':
            self.registration.SetMetricAsMeanSquares()
        else:
            raise ValueError("Unknown metric: %s" % options.registration_method)

        self.registration.SetMetricSamplingStrategy(self.registration.RANDOM)
        self.registration.SetMetricSamplingPercentage(
            options.sampling_percentage)

    def execute(self):
        """
        Run image registration. All the views are registered one by one. The
        image
        at index 0 is used as a reference.
        """

        # Get reference image.
        self.data.set_active_image(self.fixed_index,
                                   self.options.channel,
                                   self.options.scale,
                                   "original")
        fixed_image = self.data.get_itk_image()

        # Get moving image
        self.data.set_active_image(self.moving_index,
                                   self.options.channel,
                                   self.options.scale,
                                   "original")
        moving_image = self.data.get_itk_image()

        # INITIALIZATION
        # --------------
        # Start by rotating the moving image with the known rotation angle.
        print('Initializing registration')
        manual_transform = sitk.Euler3DTransform()

        # Rotate around the physical center of the image.
        rotation_center = moving_image.TransformContinuousIndexToPhysicalPoint(
            [(index - 1) / 2.0 for index in moving_image.GetSize()])
        manual_transform.SetCenter(rotation_center)

        # Rotation
        initial_rotation = self.data.get_rotation_angle(radians=True)
        if self.options.rot_axis == 0:
            manual_transform.SetRotation(initial_rotation, 0, 0)
        elif self.options.rot_axis == 1:
            manual_transform.SetRotation(0, initial_rotation, 0)
        else:
            manual_transform.SetRotation(0, 0, initial_rotation)

        # Translation
        manual_transform.SetTranslation([self.options.y_offset,
                                         self.options.x_offset,
                                         self.options.z_offset])

        modified_moving_image = ops_itk.resample_image(moving_image,
                                                       manual_transform)

        # 2. Run Automatic initialization

        transform = sitk.CenteredTransformInitializer(
            fixed_image,
            modified_moving_image,
            sitk.AffineTransform(3),
            sitk.CenteredTransformInitializerFilter.MOMENTS
        )

        # print "The initial transform is:"
        # print transform

        # Set initial transform
        self.registration.SetInitialTransform(transform)

        # SPATIAL MASK
        # =====================================================================
        # The registration metric works more reliably when it knows where
        # non-zero
        # voxels are located.
        thd = self.options.mask_threshold
        fixed_mask = sitk.BinaryDilate(
            sitk.BinaryThreshold(fixed_image, 0, thd, 0, 1))
        moving_mask = sitk.BinaryDilate(
            sitk.BinaryThreshold(modified_moving_image, 0, thd, 0, 1))

        self.registration.SetMetricFixedMask(fixed_mask)
        self.registration.SetMetricMovingMask(moving_mask)

        # START
        # ======================================================================

        if self.options.reg_enable_observers:
            # OBSERVERS
            self.registration.AddCommand(sitk.sitkStartEvent, registration.start_plot)
            self.registration.AddCommand(sitk.sitkIterationEvent, lambda: registration.plot_values(self.registration))

        print("Starting registration of views " \
              "%i (fixed) & %i (moving)" % (self.fixed_index, self.moving_index))

        result = self.registration.Execute(
            sitk.Cast(fixed_image, sitk.sitkFloat32),
            sitk.Cast(modified_moving_image, sitk.sitkFloat32))

        result = sitk.AffineTransform(result)
        # RESULTS
        # =====================================================================
        # Combine two partial transforms into one.
        # self.final_transform = sitk.Transform(manual_transform)
        # self.final_transform.AddTransform(result)

        # The two resulting transforms are combined into one here, because
        # it is easier to save a single transform into a HDF5 file.

        A0 = numpy.asarray(manual_transform.GetMatrix()).reshape(3, 3)
        c0 = numpy.asarray(manual_transform.GetCenter())
        t0 = numpy.asarray(manual_transform.GetTranslation())

        A1 = numpy.asarray(result.GetMatrix()).reshape(3, 3)
        c1 = numpy.asarray(result.GetCenter())
        t1 = numpy.asarray(result.GetTranslation())

        combined_mat = numpy.dot(A0, A1)
        combined_center = c1
        combined_translation = numpy.dot(A0, t1 + c1 - c0) + t0 + c0 - c1
        self.final_transform = sitk.AffineTransform(combined_mat.flatten(),
                                                    combined_translation,
                                                    combined_center)

        # Print final metric value and stopping condition
        print((
            'Final metric value: {0}'.format(self.registration.GetMetricValue())))
        print((
            'Optimizer\'s stopping condition, {0}'.format(
                self.registration.GetOptimizerStopConditionDescription())))
        print(self.final_transform)

        if self.options.reg_enable_observers:
            registration.end_plot(fixed_image, moving_image, self.final_transform)

    def set_moving_image(self, index):
        """
        Parameters
        ----------
        :param index    The moving image index from 0 to views number - 1

        """
        self.moving_index = index

    def set_fixed_image(self, index):
        """
        Parameters
        ----------
        :param index    The fixed image index from 0 to views number - 1. Should
                        be zero in most cases.

        """
        self.fixed_index = index

    def get_final_transform(self):
        """"
        Returns
        -------

        Get the final transform as an ITK transform
        """
        return self.final_transform

    def get_resampled_result(self):
        """

        Returns
        -------

        Get the registration result as a resampled image.
        """

        self.data.set_active_image(self.fixed_index,
                                   self.options.channel,
                                   self.options.scale,
                                   "original")

        fixed_image = self.data.get_itk_image()

        self.data.set_active_image(self.moving_index,
                                   self.options.channel,
                                   self.options.scale,
                                   "original")

        moving_image = self.data.get_itk_image()

        return ops_itk.resample_image(moving_image, self.final_transform,
                                      fixed_image)

    def save_result(self):
        scale = self.options.scale
        channel = self.options.channel
        view = self.moving_index

        # Add registered image
        self.data.set_active_image(view, channel, scale, "original")
        angle = self.data.get_rotation_angle(radians=False)
        spacing = self.data.get_voxel_size()
        registered_image = ops_itk.convert_from_itk_image(
            self.get_resampled_result())
        self.data.add_registered_image(registered_image, scale, view, channel,
                                       angle, spacing)
        # Add transform
        transform = self.get_final_transform()
        tfm_params = ops_itk.get_itk_transform_parameters(transform)
        self.data.add_transform(scale, view, channel, tfm_params[1],
                                tfm_params[2], tfm_params[0])

    def add_observers(self, start, update):
        """

        Parameters
        ----------
        start       Observer to add for the registration start
        update      Observer to add for registration progress updates.

        Returns
        -------

        """
        self.registration.AddCommand(sitk.sitkStartEvent, start)
        self.registration.AddCommand(sitk.sitkIterationEvent,
                                     lambda: update(self.registration))


class MultiViewRegistrationISM(RotatedMultiViewRegistration):

    def execute(self):
        # Get reference image.
        self.data.set_active_image(self.fixed_index,
                                   self.options.channel,
                                   self.options.scale,
                                   "original")
        fixed_image = self.data.get_itk_image()

        for idx in range(self.data.get_number_of_images("original")):
            print("Registering view {}".format(idx))
            self.moving_index = idx

            # Get moving image
            self.data.set_active_image(self.moving_index,
                                       self.options.channel,
                                       self.options.scale,
                                       "original")
            moving_image = self.data.get_itk_image()

            # Register
            self.final_transform = registration.itk_registration_rigid_2d(fixed_image,
                                                                          moving_image,
                                                                          self.options)
            self.save_result()

        print("All views registered and saved to the data structure")



import os

import matplotlib.pyplot as plt
plt.style.use("seaborn-colorblind")

import numpy as np
import miplib.processing.ndarray as arrayops

from miplib.data.containers.fourier_correlation_data import FourierCorrelationData, FourierCorrelationDataCollection
from miplib.processing.converters import degrees_to_radians


def plot_resolution_curves(data_to_plot, x_idx=0, size=(2,2), disable_ax_labels=False):
    """
    Make a figure with all FRC curves plotted into a single subplot

    :param data_to_plot: a FourierCorrelationDataCollection object with the FRC results
    :param x_idx:        if the data contains FRC results with different dynamic range (pixel size), indicate
                         the index of the dataset here with the maximum range
    :param size:        size of the plot
    :param disable_ax_labels: disable y- and x-axis labels (Correlation, Frequency), in case you want to add them
                        later for instance
    :return:            returns the matplotlib.pyplot.Figure object that you can use to make further modificaitons
                        to the plot
    """

    assert isinstance(data_to_plot, FourierCorrelationDataCollection)

    fig, ax = plt.subplots(figsize=size)

    resolution_curves_subplot(ax, data_to_plot, x_idx, disable_ax_labels)

    return fig


def resolution_curves_subplot(ax, data_to_plot, x_idx=0, disable_ax_labels=False):
    """
    Does the actual owrk fo the plot_resolution_curves function, but requires an axis as an input. It is useful eg.
    when one desires to have several subplots.

    :param ax:           plot axis (as in matplotlib subplot instance)
    :param data_to_plot: a FourierCorrelationDataCollection object with the FRC results
    :param x_idx:        if the data contains FRC results with different dynamic range (pixel size), indicate
                         the index of the dataset here with the maximum range
    :param size:        size of the plot
    :param disable_ax_labels: disable y- and x-axis labels (Correlation, Frequency), in case you want to add them
                        later for instance
    :return:            returns the matplotlib.pyplot.Figure object that you can use to make further modificaitons
                        to the plot
        """
    assert isinstance(data_to_plot, FourierCorrelationDataCollection)

    angles = list()
    datasets = list()

    # Sort datasets by angle.
    for dataset in data_to_plot:
        angles.append((int(dataset[0])))
        datasets.append(dataset[1])

    angles, datasets = list(zip(*sorted(zip(angles, datasets))))

    # plot threshold
    dataset = datasets[int(x_idx)]

    y = dataset.resolution["threshold"]
    x = dataset.correlation["frequency"]
    if x[-1] < 1.0:
        x = np.append(x, 1.0)
        y = np.append(y, y[-1])

    x_axis = arrayops.safe_divide(x, 2 * dataset.resolution["spacing"])

    ax.plot(x_axis, y, linestyle='--', color='#b5b5b3')

    if not disable_ax_labels:
        xlabel = r'Frequency ($\mathrm{\mu m}^{-1}$)'
        ylabel = 'Correlation'
        ax.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)

    for idx, dataset in enumerate(datasets):

        ax.set_ylim([0, 1.2])

        # Plot calculated FRC values as xy scatter.
        y = dataset.correlation["curve-fit"]
        x = dataset.correlation["frequency"]
        x_axis = arrayops.safe_divide(x, 2 * dataset.resolution["spacing"])

        ax.plot(x_axis, y)

    return ax


class FourierDataPlotter(object):
    """
    An attempt of sorts to make a class to handle the various types of FRC/FSC plots. I'm not quite sure
    how much sense that makes. It might be better to just have individual funcitons, more Python.
    """
    def __init__(self, data, path=None):
        assert isinstance(data, FourierCorrelationDataCollection)

        self.data = data

        if len(self.data) < 3:
            self._columns = len(self.data)
        else:
            self._columns = 3

        if len(self.data) % self._columns == 0:
            self._rows = int(len(self.data) / self._columns)
        else:
            self._rows = int(len(self.data) / self._columns + 1)

        if path is not None:
            assert os.path.isdir(path)
            self.path = path

    def plot_all(self, save_fig=False, custom_titles=None, show=True):
        """
        Plot all curves in the FourierCorrelationDataCollection data object
        that was supplied to the constructor.

        """
        axescolor = '#f6f6f6'

        size = (6, self._rows * 2) if save_fig else (12, self._rows * 4)

        fig, plots = plt.subplots(self._rows, self._columns,
                                  figsize=size)
        # rect = fig.patch
        # rect.set_facecolor('white')

        fig.tight_layout(pad=0.4, w_pad=2, h_pad=6)

        angles = list()
        datasets = list()

        # Sort datasets by angle.
        for dataset in self.data:
            angles.append((int(dataset[0])))
            datasets.append(dataset[1])

        angles, datasets = list(zip(*sorted(zip(angles, datasets))))

        if custom_titles is None:
            titles = list("FRC @ angle %i" % angle for angle in angles)
        else:
            assert len(custom_titles) == len(angles)
            titles = custom_titles

        # Make subplots
        for title, dataset, plot in zip(titles, datasets, plots.flatten()):
            self.__make_frc_subplot(plot, dataset, title)

        if save_fig:
            file_name = os.path.join(self.path, "all_frc_curves.eps")
            plt.savefig(file_name, dpi=1200)

        if show:
            plt.show()

    def plot_all_to_files(self, custom_titles=None, size=(3.3, 3), header=False):
        assert self.path is not None

        fig, plot = plt.subplots(1, 1, figsize=size, tight_layout=True)
        # plot.set(aspect='equal')

        angles = list()
        datasets = list()

        # Sort datasets by angle.
        for dataset in self.data:
            angles.append((int(dataset[0])))
            datasets.append(dataset[1])

        angles, datasets = list(zip(*sorted(zip(angles, datasets))))

        if custom_titles is None:
            titles = list("FRC @ angle %i" % angle for angle in angles)
        else:
            assert len(custom_titles) == len(angles)
            titles = custom_titles

        # Make subplots
        for title, dataset in zip(titles, datasets):
            self.__make_printable_frc_subplot(plot, dataset, title=(title if header else None))
            file_name = os.path.join(self.path, "{}.eps".format(title))
            plt.savefig(file_name, dpi=1200)
            plt.cla()

    def plot_one(self, angle):

        plt.figure(figsize=(5, 4))
        ax = plt.subplot(111)

        self.__make_frc_subplot(ax, self.data[int(angle)], "FRC at angle %s" % str(angle))

        plt.show()

    def plot_one_to_file(self, angle, filename, title=None, size=(2, 2), coerce_ticks=True, legend=False):
        fig, ax = plt.subplots(1, 1, figsize=size)

        self.__make_printable_frc_subplot(ax, self.data[int(angle)], title, coerce_ticks=coerce_ticks)
        file_name = os.path.join(self.path, "{}.eps".format(filename))
        if legend:
            fig.legend(('FRC', 'curve-fit', 'threshold', 'resolution-point'), bbox_to_anchor=(1.10, 1.0), loc=2,
                       borderaxespad=0.)

        plt.savefig(file_name, dpi=1200, bbox_inches='tight', pad_inches=0, transparent=True)

    def plot_polar(self):
        """
        Show the resolution as a 2D polar plot in which the resolution values are plotted
        as a function of rotatino angle.
        """

        angles = list()
        radii = list()

        for dataset in self.data:
            angles.append(degrees_to_radians(float(dataset[0])))
            radii.append(dataset[1].resolution["resolution"])

        angles, radii = list(zip(*sorted(zip(angles, radii))))
        angles = list(angles)
        radii = list(radii)
        angles.append(angles[0])
        radii.append(radii[0])

        radii_norm = list(i / max(radii) for i in radii)
        fig = plt.figure(figsize=(4, 4))
        ax = plt.subplot(111, projection="polar")
        ax.plot(angles, radii_norm, color='#61a2da')
        ax.set_rmax(1.2)
        r_ticks = np.linspace(0.1, 1.0, 5)
        r_ticks_scale = r_ticks * max(radii)

        x_labels = ['%.2f' % n for n in r_ticks_scale]

        ax.set_rticks(r_ticks)
        ax.set_yticklabels(x_labels)
        ax.set_rlabel_position(-80)  # get radial labels away from plotted line
        # ax.grid(True)

        # ax.set_title("The image resolution as a function of rotation angle")

        # ax.set_xlabel("XY")
        # ax.set_ylabel("Z")

        return ax

    def plot_polar_to_file(self, filename, size=(2, 2)):
        """
        Show the resolution as a 2D polar plot in which the resolution values are plotted
        as a function of rotatino angle.
        """

        angles = list()
        radii = list()

        for dataset in self.data:
            angles.append(degrees_to_radians(float(dataset[0])))
            radii.append(dataset[1].resolution["resolution"])

        angles, radii = list(zip(*sorted(zip(angles, radii))))
        angles = list(angles)
        radii = list(radii)
        angles.append(angles[0])
        radii.append(radii[0])

        radii_norm = list(i / max(radii) for i in radii)
        plt.figure(figsize=size)
        ax = plt.subplot(111, projection="polar")
        ax.plot(angles, radii_norm, color='#61a2da')
        ax.set_rmax(1.2)
        r_ticks = np.linspace(0.1, 1.0, 5)
        r_ticks_scale = r_ticks * max(radii)

        print(r_ticks_scale)
        print(max(radii))

        x_labels = ['%.2f' % n for n in r_ticks_scale]

        print(x_labels)
        ax.set_rticks(r_ticks)
        ax.set_yticklabels(x_labels)
        ax.set_rlabel_position(-80)  # get radial labels away from plotted line
        # ax.grid(True)

        # ax.set_title("The image resolution as a function of rotation angle")

        # ax.set_xlabel("XY")
        # ax.set_ylabel("Z")

        file_name = os.path.join(self.path, "{}.eps".format(filename))

        plt.savefig(file_name, dpi=1200, bbox_inches='tight', pad_inches=0, transparent=True)

    @staticmethod
    def __make_frc_subplot(ax, frc, title):
        """
        Creates a plot of the FRC curves in the curve_list. Single or multiple vurves can
        be plotted.
        """
        assert isinstance(frc, FourierCorrelationData)

        # # Font setting
        # font0 = FontProperties()
        # font1 = font0.copy()
        # font1.set_size('medium')
        # font = font1.copy()
        # font.set_family('sans')
        # rc('text', usetex=True)

        # Enable grid
        gridLineWidth = 0.2
        # ax.yaxis.grid(True, linewidth=gridLineWidth, linestyle='-', color='0.05')

        # Axis labelling
        xlabel = 'Frequency (1/um)'
        ylabel = 'Correlation'
        # ax.set_xlabel(xlabel, fontsize=12, position=(0.5, -0.2))
        # ax.set_ylabel(ylabel, fontsize=12, position=(0.5, 0.5))
        ax.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)

        ax.set_ylim([0, 1.2])

        # Title
        ax.set_title(title)

        # Plot calculated FRC values as xy scatter.
        y = frc.correlation["correlation"]
        x = frc.correlation["frequency"]
        x_axis = arrayops.safe_divide(x, 2 * frc.resolution["spacing"])

        ax.plot(x_axis, y, '^', markersize=6, color='#b5b5b3',
                label='FRC')

        # Plot polynomial fit as a line plot over the FRC scatter
        y = frc.correlation["curve-fit"]
        ax.plot(x_axis, y, linewidth=3, color='#61a2da',
                label='Least-squares fit')

        # Plot the resolution threshold curve
        y = frc.resolution["threshold"]
        res_crit = frc.resolution["criterion"]
        if res_crit == 'one-bit':
            label = 'One-bit curve'
        elif res_crit == 'half-bit':
            label = 'Half-bit curve'
        elif res_crit == 'fixed':
            label = 'y = %f' % y[0]
        else:
            label = "Threshold"

        if x[-1] < 1.0:
            x = np.append(x, 1.0)
            y = np.append(y, y[-1])

        x_axis = arrayops.safe_divide(x, 2 * frc.resolution["spacing"])

        ax.plot(x_axis, y, color='#d77186',
                label=label, lw=2, linestyle='--')

        # Plot resolution point
        y0 = frc.resolution["resolution-point"][0]
        x0 = frc.resolution["resolution-point"][1] / (2 * frc.resolution["spacing"])

        ax.plot(x0, y0, 'ro', markersize=8, label='Resolution point', color='#D75725')

        verts = [(x0, 0), (x0, y0)]
        xs, ys = list(zip(*verts))

        ax.plot(xs, ys, 'x--', lw=3, color='#D75725', ms=10)
        # ax.text(x0, y0 + 0.10, 'RESOL-FREQ', fontsize=12)

        resolution = "The resolution is {} um.".format(
            frc.resolution["resolution"])
        ax.text(0.5, -0.3, resolution, ha="center", fontsize=12)

        # x_axis = arrayops.safe_divide(np.linspace(0.0, 1.0, num=len(ax.get_xticklabels())),
        #                              2*frc.resolution["spacing"])

        # x_labels = map(lambda n: '%.1f' % n, x_axis)

        # ax.set_xticklabels(x_labels)

        # Add legend
        # ax.legend()

    @staticmethod
    def __make_printable_frc_subplot(ax, frc, title=None, coerce_ticks=True):
        """
        Creates a plot of the FRC curves in the curve_list. Single or multiple vurves can
        be plotted.
        """
        assert isinstance(frc, FourierCorrelationData)

        # # Font setting
        # font0 = FontProperties()
        # font1 = font0.copy()
        # font1.set_size('medium')
        # font = font1.copy()
        # font.set_family('sans')
        # rc('text', usetex=True)

        # Enable grid
        gridLineWidth = 0.2
        # ax.yaxis.grid(True, linewidth=gridLineWidth, linestyle='-', color='0.05')

        # Marker setup
        colorArray = ['blue', 'green', 'red', 'orange', 'brown', 'black', 'violet', 'pink']
        marker_array = ['^', 's', 'o', 'd', '1', 'v', '*', 'p']

        # Axis labelling
        xlabel = 'Frequency'
        ylabel = 'Correlation'
        # ax.set_xlabel(xlabel, fontsize=12, position=(0.5, -0.2))
        # ax.set_ylabel(ylabel, fontsize=12, position=(0.5, 0.5))
        ax.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)

        ax.set_ylim([0, 1.2])

        # Title
        if title is not None:
            ax.set_title(title)

        # Plot calculated FRC values as xy scatter.
        y = frc.correlation["correlation"]
        x_raw = frc.correlation["frequency"]
        x = arrayops.safe_divide(x_raw, 2 * frc.resolution["spacing"])
        ax.plot(x, y, marker_array[0], color='#b5b5b3',
                label='FRC')

        # Plot polynomial fit as a line plot over the FRC scatter
        y = frc.correlation["curve-fit"]
        ax.plot(x, y, color='#61a2da',
                label='Least-squares fit')

        # Plot the resolution threshold curve
        y = frc.resolution["threshold"]
        res_crit = frc.resolution["criterion"]
        if res_crit == 'one-bit':
            label = 'One-bit curve'
        elif res_crit == 'half-bit':
            label = 'Half-bit curve'
        elif res_crit == 'fixed':
            label = 'y = %f' % y[0]
        else:
            label = "Threshold"

        if x_raw[-1] < 1.0:
            x_th = arrayops.safe_divide(np.append(x_raw, 1.0), 2 * frc.resolution["spacing"])
            y = np.append(y, y[-1])
        else:
            x_th = x

        ax.plot(x_th, y, color='#d77186',
                label=label, linestyle='--')

        # Plot resolution point
        y0 = frc.resolution["resolution-point"][0]
        x0 = frc.resolution["resolution-point"][1] / (2 * frc.resolution["spacing"])

        ax.plot(x0, y0, 'ro', label='Resolution point', color='#D75725')

        verts = [(x0, 0), (x0, y0)]
        xs, ys = list(zip(*verts))

        ax.plot(xs, ys, 'x--', color='#D75725', ms=10)

        # x_axis = arrayops.safe_divide(np.linspace(0.0, 1.0, num=len(ax.get_xticklabels())),
        #                               2 * frc.resolution["spacing"])
        #
        # if coerce_ticks is True:
        #     x_labels = map(lambda n: '%d' % n, x_axis)
        # else:
        #     x_labels = map(lambda n: '%.1f' % n, x_axis)
        #
        # ax.set_xticklabels(x_labels)



"""
Sami Koho - 2015 - sami.koho@gmail.com

A small utility that generates a series of Gaussian blurred images from a
single base image. This utility was use to create an autofocus function test
dataset.

"""

import os
import sys

from scipy import ndimage, misc


def main():
    if len(sys.argv) < 2 or not os.path.isfile(sys.argv[1]):
        print("Please specify a path to an image file")
        sys.exit(1)

    original = misc.imread(sys.argv[1], mode='P')

    path_parts = os.path.split(sys.argv[1])
    output_dir = os.path.join(path_parts[0], "Blurred")
    extension = path_parts[1].split(".")[1]
    base_name = path_parts[1].split(".")[0]

    if not os.path.exists(output_dir):
            os.makedirs(output_dir)

    for mul in range(0, 30):
        sigma = mul * 1
        blurred = ndimage.gaussian_filter(original, sigma)

        file_name = base_name + "_Gaussian_" + str(sigma) + "." + extension
        output_path = os.path.join(output_dir, file_name)

        misc.imsave(output_path, blurred)

if __name__ == "__main__":
    main()





import numpy as np
from mayavi import mlab

from miplib.data.iterators.fourier_shell_iterators import ConicalFourierShellIterator
from miplib.data.containers.image import Image

dataset = Image(np.ones((255,255,255), dtype=np.uint8)*20, (0.05,0.05,0.05))

iterator = ConicalFourierShellIterator(dataset.shape, 6, 20)

section = iterator[(100, 106, 30, 60)]

dataset[section] = 255

mlab.contour3d(dataset)
mlab.show()



# coding=utf-8
from math import floor

import numpy as np
import miplib.processing.converters as converters


class FourierRingIterator(object):
    """
    A Fourier ring iterator class for 2D images. Calculates a 2D polar coordinate
    centered at the geometric center of the image shape.
    """
    def __init__(self, shape, d_bin):
        """
        :param shape: the volume shape
        :param d_bin: thickness of the ring in pixels
        """

        assert len(shape) == 2

        # Get bin size
        self.d_bin = d_bin
        self.ring_start = 0
        self._nbins = int(floor(shape[0] / (2 * self.d_bin)))
        # Create Fourier grid
        axes = (np.arange(-np.floor(i / 2.0), np.ceil(i / 2.0)) for i in shape)
        y, x = np.meshgrid(*axes)
        self.meshgrid = (y, x)

        # Create OP vector array
        self.r = np.sqrt(x ** 2 + y ** 2)
        # Current ring index
        self.current_ring = self.ring_start

        self.freq_nyq = int(np.floor(shape[0] / 2.0))
        self._radii = np.arange(0, self.freq_nyq, self.d_bin)

    @property
    def radii(self): return self._radii

    @property
    def nbins(self): return self._nbins

    def get_points_on_ring(self, ring_start, ring_stop):

        arr_inf = self.r >= ring_start
        arr_sup = self.r < ring_stop

        return arr_inf*arr_sup

    def __iter__(self):
        return self

    def __next__(self):
        if self.current_ring < self._nbins:
            ring = self.get_points_on_ring(self.current_ring * self.d_bin,
                                           (self.current_ring + 1) * self.d_bin)
        else:
            raise StopIteration

        self.current_ring += 1
        return np.where(ring), self.current_ring-1


class SectionedFourierRingIterator(FourierRingIterator):
    """
    An iterator for 2D images. Includes the option use only a specific rotated section of
    the fourier ring for FRC calculation.
    """
    def __init__(self, shape, d_bin, d_angle):
        """
        :param shape: Shape of the image
        :param d_bin: The radius increment size (pixels)
        :param d_angle: The angle increment size (degrees)
        """

        FourierRingIterator.__init__(self, shape, d_bin)

        self.d_angle = converters.degrees_to_radians(d_angle)

        y, x = self.meshgrid

        # Create inclination and azimuth angle arrays
        self.phi = np.arctan2(y, x) + np.pi

        self.phi += self.d_angle/2
        self.phi[self.phi >= 2*np.pi] -= 2*np.pi

        self._angle = 0
        
        self.angle_sector = self.get_angle_sector(0, d_bin)

    @property
    def angle(self):
        return self._angle
    
    @angle.setter
    def angle(self, value):
        angle = converters.degrees_to_radians(value)
        self._angle = angle
        self.angle_sector = self.get_angle_sector(angle, angle + self.d_angle)
        
    def get_angle_sector(self, phi_min, phi_max):
        """
        Use this to extract
        a section from a sphere that is defined by start and stop angles.

        :param phi_min: the angle at which to start the section, in radians
        :param phi_max: the angle at which to stop the section, in radians
        :return:

        """
        arr_inf = self.phi >= phi_min
        arr_sup = self.phi < phi_max

        arr_inf_neg = self.phi >= phi_min + np.pi
        arr_sup_neg = self.phi < phi_max + np.pi

        return arr_inf * arr_sup + arr_inf_neg * arr_sup_neg

    def __getitem__(self, limits):
        """
        Get a single conical section of a 2D ring.

        :param limits:  a list of parameters (ring_start, ring_stop, angle_min, angle_ma)
        that are required to define a single section of a fourier ring.
         """
        (ring_start, ring_stop, angle_min, angle_max) = limits
        ring = self.get_points_on_ring(ring_start, ring_stop)
        cone = self.get_angle_sector(angle_min, angle_max)

        return np.where(ring*cone)

    def __next__(self):
        if self.current_ring < self._nbins:
            ring = self.get_points_on_ring(self.current_ring * self.d_bin,
                                           (self.current_ring + 1) * self.d_bin)
        else:
            raise StopIteration

        self.current_ring += 1
        return np.where(ring*self.angle_sector), self.current_ring-1



"""
fusion.py

Copyright (C) 2014, 2016 Sami Koho
All rights reserved.

This software may be modified and distributed under the terms
of the BSD license.  See the LICENSE file for details.

This file contains the miplib multi-view image fusion algorithms.
They have been inteded for use with computers that do not support
hardware GPU acceleration. THe accelerated versions of the same functions
can be found in fusion_cuda.py. The fftconvolve function that is
used in this file, can take advantage of MKL optimizations available
in the Anaconda Accelerate package.

"""

import itertools
import os
import shutil
import sys
import tempfile
import time
import pandas

import numpy
import miplib.processing.ops_ext as ops_ext
from scipy.ndimage.interpolation import zoom
from scipy.ndimage.filters import uniform_filter
from scipy.signal import fftconvolve, medfilt

import miplib.processing.to_string as ops_output
import miplib.processing.ndarray
from miplib.data.containers import temp_data
from miplib.data.containers.image import Image
from miplib.data.messages.image_writer_wrappers import ImageWriterBase
from numpy.fft import fftn, fftshift

import miplib.psf.frc_psf as frc_psf

class DeconvolutionRL(object):
    """
    The Richardson-Lucy fusion is a result of simultaneous deblurring of
    several 3D volumes.
    """

    def __init__(self, image, psf, writer, options):
        """
        :param image:    a MyImage object

        :param options: command line options that control the behavior
                        of the fusion algorithm
        """
        assert isinstance(image, Image)
        assert isinstance(psf, Image)
        if options.save_intermediate_results:
            assert issubclass(writer.__class__, ImageWriterBase)

        self.image = image
        self.psf = psf
        self.options = options
        self.writer = writer

        self.image_size = numpy.array(self.image.shape)
        self.image_spacing = self.image.spacing
        self.psf_spacing = self.psf.spacing
        self.imdims = image.ndim

        self.__get_psfs()

        if options.verbose:
            print("The original image size is %s" % (self.image_size,))

        self.iteration_count = 0

        # Setup blocks
        self.num_blocks = options.num_blocks
        self.block_size, self.image_size = self.__calculate_block_and_image_size()
        self.memmap_directory = tempfile.mkdtemp()

        # Memmap the estimates to reduce memory requirements. This will slow
        # down the fusion process considerably..
        if self.options.memmap_estimates:
            estimate_new_f = os.path.join(self.memmap_directory, "estimate_new.dat")
            self.estimate_new = Image(numpy.memmap(estimate_new_f, dtype='float32',
                                                   mode='w+',
                                                   shape=tuple(self.image_size)), self.image_spacing)

            estimate_f = os.path.join(self.memmap_directory, "estimate.dat")
            self.estimate = Image(numpy.memmap(estimate_f, dtype=numpy.float32,
                                               mode='w+',
                                               shape=tuple(self.image_size)), self.image_spacing)
        else:
            self.estimate = Image(numpy.zeros(tuple(self.image_size),
                                              dtype=numpy.float32), self.image_spacing)
            self.estimate_new = Image(numpy.zeros(tuple(self.image_size),
                                                  dtype=numpy.float32), self.image_spacing)

        if not self.options.disable_tau1:
            prev_estimate_f = os.path.join(self.memmap_directory, "prev_estimate.dat")
            self.prev_estimate = Image(numpy.memmap(prev_estimate_f, dtype=numpy.float32,
                                                    mode='w+',
                                                    shape=tuple(self.image_size)), self.image_spacing)

        padded_block_size = tuple(i + 2 * self.options.block_pad for i in self.block_size)
        if options.verbose:
            print("The deconvolution will be run with %i blocks" % self.num_blocks)
            print("The internal block size is %s" % (padded_block_size,))

        # Create temporary directory and data file.
        self.column_headers = ('t', 'tau1', 'leak', 'e',
                             's', 'u', 'n', 'uesu')
        self._progress_parameters = numpy.empty((self.options.max_nof_iterations, len(self.column_headers)),
                                                dtype=numpy.float32)

    @property
    def progress_parameters(self):
        return pandas.DataFrame(data=self._progress_parameters, columns=self.column_headers)

    def compute_estimate(self):
        """
        Calculates a single RL deconvolution estimate. There is no reason to call this
        function -- it is used internally by the class during fusion process.
        """

        if self.options.verbose:
            print('Beginning the computation of the %i. estimate' % self.iteration_count)

        self.estimate_new[:] = numpy.float32(0)

        # Iterate over blocks
        block_nr = 1
        iterables = (range(0, m, n) for m, n in zip(self.image_size, self.block_size))
        pad = self.options.block_pad
        cache_idx = tuple(slice(pad, pad + block) for block in self.block_size)

        for idx in itertools.product(*iterables):

            estimate_idx = tuple(slice(j, j+k) for j, k in zip(idx, self.block_size))

            index = numpy.array(idx, dtype=int)

            if self.options.block_pad > 0:
                estimate_block = self.get_padded_block(
                    self.estimate, index.copy())
                image_block = self.get_padded_block(self.image, index.copy())
            else:
                estimate_block = self.estimate[estimate_idx]
                image_block = self.image[estimate_idx]

            # print "The current block is %i" % block_nr
            block_nr += 1

            # Execute: cache = convolve(PSF, estimate), non-normalized
            cache = fftconvolve(estimate_block, self.psf, mode='same')

            if self.options.rl_background != 0:
                cache += self.options.rl_background

                # ops_ext.inverse_division_inplace(cache, image_block)
            with numpy.errstate(divide="ignore"):
                cache = image_block.astype(numpy.float32) / cache
                cache[cache == numpy.inf] = 0.0
                cache = numpy.nan_to_num(cache)

            # Execute: cache = convolve(PSF(-), cache), inverse of non-normalized
            # Convolution with virtual PSFs is performed here as well, if
            # necessary
            cache = fftconvolve(cache, self.adj_psf, mode='same')

            self.estimate_new[estimate_idx] = cache[cache_idx]

        if self.options.tv_lambda > 0 and self.iteration_count > 0:
            if self.estimate.ndim == 2:
                spacing = list(self.image_spacing)
                spacing.insert(0,1)
                dv_est = ops_ext.div_unit_grad(numpy.expand_dims(self.estimate, 0),
                                               spacing)[0]
            else:
                dv_est = ops_ext.div_unit_grad(self.estimate, self.image_spacing)
            with numpy.errstate(divide="ignore"):
                self.estimate_new /= (1.0 - self.options.tv_lambda * dv_est)
                self.estimate_new[self.estimate_new == numpy.inf] = 0.0
                self.estimate_new[:] = numpy.nan_to_num(self.estimate_new)

        return ops_ext.update_estimate_poisson(self.estimate,
                                               self.estimate_new,
                                               self.options.convergence_epsilon)

    def execute(self):
        """
        This is the main fusion function
        """

        save_intermediate_results = self.options.save_intermediate_results

        first_estimate = self.options.first_estimate

        if first_estimate == 'image':
            self.estimate[:] = self.image[:].astype(numpy.float32)
        elif first_estimate == 'blurred':
            self.estimate[:] = uniform_filter(self.image, 3).astype(numpy.float32)
        elif first_estimate == 'image_mean':
            self.estimate[:] = numpy.float32(numpy.mean(self.image[:]))
        elif first_estimate == 'constant':
            self.estimate[:] = numpy.float32(self.options.estimate_constant)
        else:
            raise NotImplementedError(repr(first_estimate))

        self.iteration_count = 0
        max_count = self.options.max_nof_iterations
        initial_photon_count = self.image[:].sum()

        bar = ops_output.ProgressBar(0,
                                     max_count,
                                     totalWidth=40,
                                     show_percentage=False)

        self._progress_parameters = numpy.zeros((self.options.max_nof_iterations, len(self.column_headers)),
                                                dtype=numpy.float32)


        # duofrc_prev = 0
        # The Fusion calculation starts here
        # ====================================================================
        try:
            while True:

                if (
                        self.options.update_blind_psf > 0 and 
                        self.iteration_count > 0 and 
                        (self.iteration_count+1) % self.options.update_blind_psf == 0
                   ):
                    self.psf = frc_psf.generate_frc_based_psf(Image(self.estimate, self.image_spacing), self.options)
                    self.__get_psfs()

                info_map = {}
                ittime = time.time()

                self.prev_estimate[:] = self.estimate.copy()

                e, s, u, n = self.compute_estimate()

                self.iteration_count += 1
                photon_leak = 1.0 - (e + s + u) / initial_photon_count
                u_esu = u / (e + s + u)

                tau1 = abs(self.estimate - self.prev_estimate).sum() / abs(
                    self.prev_estimate).sum()
                info_map['TAU1=%s'] = tau1

                t = time.time() - ittime
                leak = 100 * photon_leak

                if self.options.verbose:
                    # Update UI
                    info_map['E/S/U/N=%s/%s/%s/%s'] = int(e), int(s), int(u), int(n)
                    info_map['LEAK=%s%%'] = leak
                    info_map['U/ESU=%s'] = u_esu
                    info_map['TIME=%ss'] = t

                    bar.updateComment(' ' + ', '.join([k % (ops_output.tostr(info_map[k])) for k in sorted(info_map)]))
                    bar(self.iteration_count)
                    print()

                # Save parameters to file
                self._progress_parameters[self.iteration_count - 1] = (t, tau1, leak, e, s, u, n, u_esu)


                # Save intermediate image
                if save_intermediate_results:
                    # self.temp_data.save_image(
                    #     self.estimate,
                    #     'result_%s.tif' % self.iteration_count
                    # )
                    self.writer.write(Image(self.estimate, self.image_spacing))

                # Check if it's time to stop:
                if int(u) == 0 and int(n) == 0:
                    stop_message = 'The number of non converging photons reached to zero.'
                    break
                elif self.iteration_count >= max_count:
                    stop_message = 'The number of iterations reached to maximal count: %s' % max_count
                    break
                elif not self.options.disable_tau1 and tau1 <= self.options.stop_tau:
                    stop_message = 'Desired tau-threshold achieved'
                    break
                # elif self.iteration_count >= 4 and abs(frc_diff) <= .0001:
                #     stop_message = 'FRC stop condition reached'
                #     break
                else:
                    continue

        except KeyboardInterrupt:
            stop_message = 'Iteration was interrupted by user.'

        # if self.num_blocks > 1:
        #     self.estimate = self.estimate[0:real_size[0], 0:real_size[1], 0:real_size[2]]
        if self.options.verbose:
            print()
            bar.updateComment(' ' + stop_message)
            bar(self.iteration_count)
            print()

    def __get_psfs(self):
        """
        Reads the PSFs from the HDF5 data structure and zooms to the same pixel
        size with the registered images, of selected scale and channel.
        """
        psf_orig = self.psf[:]

        # Zoom to the same voxel size
        zoom_factors = tuple(x / y for x, y in zip(self.psf_spacing, self.image_spacing))
        psf_new = zoom(psf_orig, zoom_factors).astype(numpy.float32)

        psf_new /= psf_new.sum()

        # Save the zoomed and rotated PSF, as well as its mirrored version
        self.psf = psf_new
        if self.imdims == 3:
            self.adj_psf = psf_new[::-1, ::-1, ::-1]
        else:
            self.adj_psf = psf_new[::-1, ::-1]

    def get_result(self):
        """
        Show fusion result. This is a temporary solution for now
        calling Fiji through ITK. An internal viewer would be
        preferable.
        """

        return Image(self.estimate, self.image_spacing)

    def __calculate_block_and_image_size(self):
        """
        Calculate the block size and the internal image size for a given
        number of blocks. 1,2,4 or 8 blocks are currently supported.

        """
        block_size = self.image_size
        image_size = self.image_size

        if self.num_blocks == 1:
            return block_size, image_size
        elif self.num_blocks == 2:
            multiplier3 = numpy.array([2, 1, 1])
            multiplier2 = numpy.array([2, 1])
        elif self.num_blocks == 4:
            multiplier3 = numpy.array([4, 1, 1])
            multiplier2 = numpy.array([2, 2])
        elif self.num_blocks == 8:
            multiplier3 = numpy.array([4, 2, 1])
            multiplier2 = numpy.array([4, 2])
        elif self.num_blocks == 12:
            multiplier3 = numpy.array([4, 2, 2])
            multiplier2 = numpy.array([4, 3])
        elif self.num_blocks == 24:
            multiplier3 = numpy.array([4, 3, 2])
            multiplier2 = numpy.array([6, 4])
        elif self.num_blocks == 48:
            multiplier3 = numpy.array([4, 4, 3])
            multiplier2 = numpy.array([8, 6])
        elif self.num_blocks == 64:
            multiplier3 = numpy.array([4, 4, 4])
            multiplier2 = numpy.array([8, 8])
        elif self.num_blocks == 96:
            multiplier3 = numpy.array([6, 4, 4])
            multiplier2 = numpy.array([12, 8])
        elif self.num_blocks == 144:
            multiplier3 = numpy.array([4, 6, 6])
            multiplier2 = numpy.array([12, 12])
        else:
            raise NotImplementedError

        if self.imdims == 2:
            block_size = numpy.ceil(self.image_size.astype(numpy.float16) / multiplier2).astype(numpy.int64)
            image_size += (multiplier2 * block_size - image_size)
        else:
            block_size = numpy.ceil(self.image_size.astype(numpy.float16) / multiplier3).astype(numpy.int64)
            image_size += (multiplier3 * block_size - image_size)

        return block_size, image_size

    def get_padded_block(self, image, block_start_index):
        """
        Get a padded block from the self.estimate

        Parameters
        ----------
        :param image: a numpy.ndarray or or its subclass
        :param block_start_index  The real block start index, not considering the padding

        Returns
        -------
        Returns the padded estimate block as a numpy array.

        """

        block_pad = self.options.block_pad
        image_size = self.image_size
        ndims = self.imdims

        # Apply padding
        end_index = block_start_index + self.block_size + block_pad
        start_index = block_start_index - block_pad

        idx = tuple(slice(start, stop) for start, stop in zip(start_index, end_index))

        # If the padded block fits within the image boundaries, nothing special
        # is needed to extract it. Normal numpy slicing notation is used.
        if (image_size >= end_index).all() and (start_index >= 0).all():
            return image[idx]

        else:
            block_size = tuple(i + 2 * block_pad for i in self.block_size)
            # Block outside the image boundaries will be filled with zeros.
            block = numpy.zeros(block_size)
            # If the start_index is close to the image boundaries, it is very
            # probable that padding will introduce negative start_index values.
            # In such case the first pixel index must be corrected.
            if (start_index < 0).any():
                block_start = numpy.negative(start_index.clip(max=0))
                image_start = start_index + block_start
            else:
                block_start = (0,) * ndims
                image_start = start_index

            # If the padded block is larger than the image size the
            # block_size must be adjusted.
            if not (image_size >= end_index).all():
                block_crop = end_index - image_size
                block_crop[block_crop < 0] = 0
                block_end = block_size - block_crop
            else:
                block_end = block_size

            end_index = start_index + block_end

            block_idx = tuple(slice(start, stop) for start, stop in zip(block_start, block_end))
            image_idx = tuple(slice(start, stop) for start, stop in zip(image_start, end_index))

            block[block_idx] = image[image_idx]

            return block

    def get_8bit_result(self, denoise=False):
        """
        Returns the current estimate (the fusion result) as an 8-bit uint, rescaled
        to the full 0-255 range.
        """
        if denoise:
            image = medfilt(self.estimate)
        else:
            image = self.estimate

        image *= (255.0 / image.max())
        image[image < 0] = 0
        return Image(image.astype(numpy.uint8), self.image_spacing)

    # def get_saved_data(self):
    #     return pandas.DataFrame(columns=self.column_headers, data=self._progress_parameters)

    def close(self):
        if self.options.memmap_estimates:
            del self.estimate
            del self.estimate_new
        if not self.options.disable_tau1:
            del self.prev_estimate

        shutil.rmtree(self.memmap_directory)

        #self.temp_data.close_data_file()



def isiterable(something):
    """ Check if a variable is iterable
    
    :param something: some variable that you are interested in
    :type something: any
    :return: True/False
    :rtype: boolean
    """
    try:
        iter(something)
        return True
    except TypeError:
        return False


import numpy
from functools import reduce

def nroot(array, n):
    """

    :param array:   A n dimensional numpy array by default. Of course this works
                    with single numbers and whatever the interpreter can understand
    :param n:       The root - a number
    :return:
    """
    return array ** (1.0 / n)


def normalize(array):
    """
    Normalizes a numpy array by dividing each element with the array.sum()

    :param array: a numpy.array
    :return:
    """
    return array / array.sum()


def float2dtype(float_type):
    """Return numpy float dtype object from float type label.
    """
    if float_type == 'single' or float_type is None:
        return numpy.float32
    if float_type == 'double':
        return numpy.float64
    raise NotImplementedError (repr(float_type))


def contract_to_shape(data, shape):
    """
    Remove padding from input data array. The function
    expects the padding to be symmetric on all sides
    """
    assert all(x <= y for x,y in zip(shape, data.shape))

    if any(x != y for x,y in zip(shape, data.shape)):

        slices = []
        for s1, s2 in zip(data.shape, shape):
            slices.append(slice((s1 - s2) // 2, (s1 + s2) // 2))

        image = data[tuple(slices)]
    else:
        image = data

    return image

def expand_to_shape(data, shape, dtype=None, background=None):
    """
    Expand data to given shape by zero-padding.
    """
    if dtype is None:
        dtype = data.dtype

    start_index = numpy.array(shape) - data.shape
    data_start = numpy.negative(start_index.clip(max=0))
    data = cast_to_dtype(data, dtype, rescale=False)
    if data.ndim == 3:
        data = data[data_start[0]:, data_start[1]:, data_start[2]:]
    else:
        data = data[data_start[0]:, data_start[1]:]

    if background is None:
        background = 0

    if (shape != data.shape):
        expanded_data = numpy.zeros(shape, dtype=dtype) + background
        slices = []
        rhs_slices = []
        for s1, s2 in zip(shape, data.shape):
            a, b = (s1 - s2 + 1) // 2, (s1 + s2 + 1) // 2
            c, d = 0, s2
            while a < 0:
                a += 1
                b -= 1
                c += 1
                d -= 1
            slices.append(slice(a, b))
            rhs_slices.append(slice(c, d))
        try:
            expanded_data[tuple(slices)] = data[tuple(rhs_slices)]
        except ValueError:
            print(data.shape, shape)
            raise
        return expanded_data
    else:
        return data


def mul_seq(seq):
    return reduce(lambda x, y: x * y, seq, 1)


def float2dtype(float_type):
    """Return numpy float dtype object from float type label.
    """
    if float_type == 'single' or float_type is None:
        return numpy.float32
    if float_type == 'double':
        return numpy.float64
    raise NotImplementedError(repr(float_type))

def cast_to_dtype(data, dtype, rescale=True, remove_outliers=False):
    """
     A function for casting a numpy array into a new data type.
    The .astype() property of Numpy sometimes produces satisfactory
    results, but if the data type to cast into has a more limited
    dynamic range than the original data type, problems may occur.

    :param data:            a numpy.array object
    :param dtype:           data type string, as in Python
    :param rescale:         switch to enable rescaling pixel
                            values to the new dynamic range.
                            This should always be enabled when
                            scaling to a more limited range,
                            e.g. from float to int
    :param remove_outliers: sometimes deconvolution/fusion generates
                            bright artifacts, which interfere with
                            the rescaling calculation. You can remove them
                            with this switch
    :return:                Returns the input data, cast into the new datatype
    """
    if data.dtype == dtype:
        return data

    if 'int' in str(dtype):
        data_info = numpy.iinfo(dtype)
        data_max = data_info.max
        data_min = data_info.min
    elif 'float' in str(dtype):
        data_info = numpy.finfo(dtype)
        data_max = data_info.max
        data_min = data_info.min
    else:
        data_max = data.max()
        data_min = data.min()
        print("Warning casting into unknown data type. Detail clipping" \
              "may occur")

    # In case of unsigned integers, numbers below zero need to be clipped
    if 'uint' in str(dtype):
        data_max = 255
        data_min = 0

    if remove_outliers:
        data = data.clip(0, numpy.percentile(data, 99.99))

    if rescale is True:
        return rescale_to_min_max(data, data_min, data_max).astype(dtype)
    else:
        return data.clip(data_min, data_max).astype(dtype)


def rescale_to_min_max(data, data_min, data_max):
    """
    A function to rescale image intensities to range, define by
    data_min and data_max input parameters.

    :param data:        Input image (Numpy array)
    :param data_min:    Minimum pixel value. Can be any type of a number
                        (preferably of the same type with the data.dtype)
    :param data_max:    Maximum pixel value
    :return:            Return the rescaled array
    """
    # Return array with max value in the original image scaled to correct
    # range
    if abs(data.max()) > abs(data.min()) or data_min == 0:
        return data_max / data.max() * data
    else:
        return data_min / data.min() * data

def safe_divide(numerator, denominator):
    """
    Division of numpy arrays that can handle division by zero. NaN results are
    coerced to zero. Also suppresses the division by zero warning.
    :param numerator:
    :param denominator:
    :return:
    """
    with numpy.errstate(divide="ignore"):
        result = numerator / denominator
        result[result == numpy.inf] = 0.0
        return numpy.nan_to_num(result)


def start_to_stop_idx(start, stop):
    """
    Generate n-dimensional indexing strucure for a numpy array,
    consisting of a start-to-stop slice in each dimension
    :param start: start indexes
    :param stop: stop indexes
    :return:
    """
    return tuple(slice(a, b) for a, b in zip(start, stop))


def start_to_offset_idx(start, offset):
    """
    Generate n-dimensional indexing structure for a numpy array,
    based on start indexes and offsets
    :param start: list of indexes to start the slicing from
    :param offset: list of slice lengths
    :return:
    """
    stop = start + offset
    return tuple(slice(a, b) for a, b in zip(start, stop))


def reverse_array(array):

    assert isinstance(array, numpy.ndarray)

    temp = array.copy()
    for i in range(temp.ndim):
        temp = numpy.flip(temp, i)

    return temp


# coding=utf-8
from math import floor

import numpy as np

import miplib.processing.converters as converters
import miplib.processing.ndarray as nputils
import miplib.processing.itk as itkutils


class FourierShellIterator(object):
    """
    A Simple Fourier Shell Iterator. Basically the same as a Fourier Ring Iterator,
    but for 3D.
    """

    def __init__(self, shape, d_bin):
        self.d_bin = d_bin

        # Create Fourier grid
        axes = (np.arange(-np.floor(i / 2.0), np.ceil(i / 2.0)) for i in shape)
        z, y, x = np.meshgrid(*axes)
        self.meshgrid = (z, y, x)

        # Create OP vector array
        self.r = np.sqrt(x ** 2 + y ** 2 + z ** 2)

        self.shell_start = 0
        self.shell_stop = int(floor(shape[0] / (2 * self.d_bin))) - 1

        self.current_shell = self.shell_start

        self.freq_nyq = int(np.floor(shape[0] / 2.0))

        self.radii = np.arange(0, self.freq_nyq, self.d_bin)

    @property
    def steps(self):
        return self.radii

    @property
    def nyquist(self):
        return self.freq_nyq

    def get_points_on_shell(self, shell_start, shell_stop):

        arr_inf = self.r >= shell_start
        arr_sup = self.r < shell_stop

        return arr_inf*arr_sup

    def __getitem__(self, limits):
        """
        Get a points on a Fourier shell specified by the start and stop coordinates

        :param shell_start: The start of the shell (0 ... Nyquist)
        :param shell_stop:  The end of the shell

        :return:            Returns the coordinates of the points that are located on
                            the specified shell
        """
        (shell_start, shell_stop) = limits
        shell = self.get_points_on_shell(shell_start, shell_stop)
        return np.where(shell)

    def __iter__(self):
        return self

    def __next__(self):

        shell_idx = self.current_shell

        if shell_idx <= self.shell_stop:
            shell = self.get_points_on_shell(self.current_shell * self.d_bin,
                                             (self.current_shell + 1) * self.d_bin)
        else:
            raise StopIteration

        self.current_shell += 1

        return np.where(shell), shell_idx


class ConicalFourierShellIterator(FourierShellIterator):
    """
    An iterator for 3D images. Includes the option section a single shell into rotational
    sections.
    """
    def __init__(self, shape, d_bin, d_angle):
        """
        :param shape: Shape of the image
        :param d_bin: The radius increment size (pixels)
        :param d_angle: The angle increment size (degrees)
        """

        FourierShellIterator.__init__(self, shape, d_bin)

        self.d_angle = converters.degrees_to_radians(d_angle)

        z, y, x = self.meshgrid

        # Create inclination and azimuth angle arrays
        self.phi = np.arctan2(y, z) + np.pi

        self.phi += self.d_angle/2
        self.phi[self.phi >= 2*np.pi] -= 2*np.pi

        self.rotation_start = 0
        self.rotation_stop = 360 / d_angle - 1
        self.current_rotation = self.rotation_start

        self.angles = np.arange(0, 360, d_angle, dtype=int)

    @property
    def steps(self):
        return self.radii, self.angles

    def get_angle_sector(self, phi_min, phi_max):
        """
        Assuming a classical spherical coordinate system the azimutahl
        angle is the angle between the x- and y- axes. Use this to extract
        a conical section from a sphere that is defined by start and stop azimuth
        angles.

        :param phi_min: the angle at which to start the section, in radians
        :param phi_max: the angle at which to stop the section, in radians
        :return:

        """
        arr_inf = self.phi >= phi_min
        arr_sup = self.phi < phi_max

        arr_inf_neg = self.phi >= phi_min + np.pi
        arr_sup_neg = self.phi < phi_max + np.pi

        return arr_inf * arr_sup + arr_inf_neg * arr_sup_neg

    def __getitem__(self, limits):
        """
        Get a single conical section of a 3D shell.

        :param shell_start: The start of the shell (0 ... Nyquist)
        :param shell_stop:  The end of the shell
        :param angle_min:   The start of the cone (degrees 0-360)
        :param angle_max:   The end of the cone
        :return:            Returns the coordinates of the points that are located inside
                            the portion of a shell that intersects with the points on the
                            cone.
        """
        (shell_start, shell_stop, angle_min, angle_max) = limits
        angle_min = converters.degrees_to_radians(angle_min)
        angle_max = converters.degrees_to_radians(angle_max)

        shell = self.get_points_on_shell(shell_start, shell_stop)
        cone = self.get_angle_sector(angle_min, angle_max)

        return np.where(shell*cone)

    def __next__(self):

        rotation_idx = self.current_rotation
        shell_idx = self.current_shell

        if rotation_idx <= self.rotation_stop and shell_idx <= self.shell_stop:
            shell = self.get_points_on_shell(self.current_shell * self.d_bin,
                                             (self.current_shell + 1) * self.d_bin)

            cone = self.get_angle_sector(self.current_rotation * self.d_angle,
                                          (self.current_rotation + 1) * self.d_angle)
        else:
            raise StopIteration

        if rotation_idx >= self.rotation_stop:
            self.current_rotation = 0
            self.current_shell += 1
        else:
            self.current_rotation += 1

        return np.where(shell*cone), shell_idx, rotation_idx


class HollowConicalFourierShellIterator(ConicalFourierShellIterator):
    """
    A conical Fourier shell iterator with the added possibility to remove
    a central section of the cone, to better deal with interpolation artefacts etc.
    """

    def __init__(self,  shape, d_bin, d_angle, d_extract_angle=5):

        ConicalFourierShellIterator.__init__(self, shape, d_bin, d_angle)

        self.d_extract_angle = converters.degrees_to_radians(d_extract_angle)

    def get_angle_sector(self, phi_min, phi_max):
        """
        Assuming a classical spherical coordinate system the azimutahl
        angle is the angle between the x- and y- axes. Use this to extract
        a conical section from a sphere that is defined by start and stop azimuth
        angles.

        In the hollow implementation a small slice in the center of the section is
        removed to avoid the effect of resampling when calculating the resolution
        along the lowest resolution axis (z), on images with very isotropic resolution
        (e.g. STED).

        :param phi_min: the angle at which to start the section, in radians
        :param phi_max: the angle at which to stop the section, in radians
        :return:

        """
        # Calculate angular sector
        arr_inf = self.phi >= phi_min
        arr_sup = self.phi < phi_max

        arr_inf_neg = self.phi >= phi_min + np.pi
        arr_sup_neg = self.phi < phi_max + np.pi

        full_section = arr_inf * arr_sup + arr_inf_neg * arr_sup_neg

        # Calculate part of the section to exclude
        sector_center = phi_min + (phi_max-phi_min)/2
        phi_min_ext = sector_center - self.d_extract_angle
        phi_max_ext = sector_center + self.d_extract_angle

        arr_inf_ext = self.phi >= phi_min_ext
        arr_sup_ext = self.phi < phi_max_ext

        arr_inf_neg_ext = self.phi >= phi_min_ext + np.pi
        arr_sup_neg_ext = self.phi < phi_max_ext + np.pi

        extract_section = arr_inf_ext * arr_sup_ext + arr_inf_neg_ext * arr_sup_neg_ext

        return full_section - extract_section


class AxialExcludeHollowConicalFourierShellIterator(HollowConicalFourierShellIterator):
    """
    A conical Fourier shell iterator with the added possibility to remove
    a central section of the cone, to better deal with interpolation artefacts etc.
    """

    def __init__(self,  shape, d_bin, d_angle, d_extract_angle=5):

        HollowConicalFourierShellIterator.__init__(self, shape, d_bin, d_angle)


    def get_angle_sector(self, phi_min, phi_max):
        """
        Assuming a classical spherical coordinate system the azimutahl
        angle is the angle between the x- and y- axes. Use this to extract
        a conical section from a sphere that is defined by start and stop azimuth
        angles.

        In the hollow implementation a small slice in the center of the section is
        removed to avoid the effect of resampling when calculating the resolution
        along the lowest resolution axis (z), on images with very isotropic resolution
        (e.g. STED).

        :param phi_min: the angle at which to start the section, in radians
        :param phi_max: the angle at which to stop the section, in radians
        :return:

        """
        # Calculate angular sector
        arr_inf = self.phi >= phi_min
        arr_sup = self.phi < phi_max

        arr_inf_neg = self.phi >= phi_min + np.pi
        arr_sup_neg = self.phi < phi_max + np.pi

        full_section = arr_inf * arr_sup + arr_inf_neg * arr_sup_neg

        axis_pos = converters.degrees_to_radians(90) + self.d_angle/2
        axis_neg = converters.degrees_to_radians(270) + self.d_angle/2

        if phi_min <= axis_pos <= phi_max:
            phi_min_ext = axis_pos - self.d_extract_angle
            phi_max_ext = axis_pos + self.d_extract_angle

        elif phi_min <= axis_neg <= phi_max:

            # Calculate part of the section to exclude
            phi_min_ext = axis_neg - self.d_extract_angle
            phi_max_ext = axis_neg + self.d_extract_angle

        else:
            return full_section

        arr_inf_ext = self.phi >= phi_min_ext
        arr_sup_ext = self.phi < phi_max_ext

        arr_inf_neg_ext = self.phi >= phi_min_ext + np.pi
        arr_sup_neg_ext = self.phi < phi_max_ext + np.pi

        extract_section = arr_inf_ext * arr_sup_ext + arr_inf_neg_ext * arr_sup_neg_ext

        return full_section - extract_section


class RotatingFourierShellIterator(FourierShellIterator):
    """
    A 3D Fourier Ring Iterator -- not a Fourier Shell Iterator, but rather
    single planes are extracted from a 3D shape by rotating the XY plane,
    as in:

    Nieuwenhuizen, Rpj, K. A. Lidke, and Mark Bates. 2013.
    “Measuring Image Resolution in Optical Nanoscopy.” Nature
    advance on (April). https://doi.org/10.1038/nmeth.2448.

    Here the shell iteration is still in 3D (for compatilibility with the others
    which doesn't make much sense in terms of calculation effort, but it should make
    it possible to

    """

    def __init__(self, shape, d_bin, d_angle):
        """
        :param shape: Shape of the image
        :param d_bin: The radius increment size (pixels)
        :param d_angle: The angle increment size (degrees)
        """

        assert len(shape) == 3, "This iterator assumes a 3D shape"

        FourierShellIterator.__init__(self, shape, d_bin)

        plane = nputils.expand_to_shape(np.ones((1, shape[1], shape[2])), shape)

        self.plane = itkutils.convert_from_numpy(
            plane,
            (1, 1, 1))

        self.rotated_plane = plane > 0

        self.rotation_start = 0
        self.rotation_stop = 360 / d_angle - 1
        self.current_rotation = self.rotation_start

        self.angles = np.arange(0, 360, d_angle, dtype=int)

    @property
    def steps(self):
        return self.radii, self.angles

    def __getitem__(self, limits):
        """
        Get a single conical section of a 3D shell.

        :param shell_start: The start of the shell (0 ... Nyquist)
        :param shell_stop:  The end of the shell
        :param angle:
        """
        (shell_start, shell_stop, angle) = limits
        rotated_plane = itkutils.convert_from_itk_image(
            itkutils.rotate_image(self.plane, angle))

        points_on_plane = rotated_plane > 0
        points_on_shell = self.get_points_on_shell(shell_start, shell_stop)

        return np.where(points_on_plane * points_on_shell)

    def __next__(self):

        rotation_idx = self.current_rotation + 1
        shell_idx = self.current_shell

        if shell_idx <= self.shell_stop:
            shell = self.get_points_on_shell(self.current_shell * self.d_bin,
                                             (self.current_shell + 1) * self.d_bin)
            self.current_shell += 1

        elif rotation_idx <= self.rotation_stop:

            rotated_plane = itkutils.convert_from_itk_image(
                itkutils.rotate_image(self.plane, self.angles[rotation_idx],
                                      interpolation='linear'))

            self.rotated_plane = rotated_plane > 0
            self.current_shell = 0
            shell_idx = 0
            self.current_rotation += 1

            shell = self.get_points_on_shell(self.current_shell * self.d_bin,
                                             (self.current_shell + 1) * self.d_bin)

        else:
            raise StopIteration

        return np.where(shell * self.rotated_plane), shell_idx, self.current_rotation





import itertools

import numpy as np
import matplotlib.pyplot as plt

from miplib.data.containers.image_data import ImageData

plt.style.use("seaborn-paper")

from miplib.data.containers.array_detector_data import ArrayDetectorData


def make_template_image(data, imagesz=250):
    """
    Makes the "fingerprint" map of the array detector images
    :param data: ArrayDetecorData object with all the iamges
    :param imagesz: integer Size of the images in pixels
    :return: returns a tuple of images (one for each channel)
    """
    assert isinstance(data, ArrayDetectorData)

    blocksz = int(imagesz/np.sqrt(data.ndetectors))

    # First calculate the total photon count for images from each detector
    # and photosensor
    pixels = np.zeros(data.ndetectors*data.ngates)

    data.iteration_axis = 'detectors'
    for idx, image in enumerate(data):
        pixels[idx] = image.sum()

    # Then generate a template image for each photosensor
    container = []
    for gate in range(data.ngates):
        image = np.zeros((imagesz, imagesz))
        idx = 0
        for x, y in itertools.product(range(0, imagesz, blocksz), range(0, imagesz, blocksz)):
            pixel_index = gate*data.ndetectors + idx
            image[x:x + blocksz, y:y + blocksz] = pixels[pixel_index]
            idx += 1

        container.append(image)

    return container


def make_psf_plot(data, size=(5,5)):
    """
    Makes a 5x5 matrix plot of Point-Spread-Functions (PSFs) that are used in the
    blind multi-image APR-ISM image fusion
    :param data: ArrayDetectorData or ImageData object that contains the 25 psfs
    :param size: Size of the plot
    :return:     Returns the figure
    """
    if isinstance(data, ArrayDetectorData):
        hdf = False
    elif isinstance(data, ImageData):
        hdf = True
    else:
        raise ValueError("data needs to be ArrayDetecorData or ImageData object")

    fig, axs = plt.subplots(5, 5, figsize=size)

    for idx, ax in enumerate(axs.flatten()):
        if hdf:
            data.set_active_image(idx, 0, 100, "psf")
            ax.imshow(data[:], cmap="hot")
        else:
            ax.imshow(data[0,idx])

        ax.get_xaxis().set_visible(False)
        ax.get_yaxis().set_visible(False)

    plt.tight_layout(pad=-0.3, w_pad=-0.3, h_pad=-0.3)

    return fig






# -*- coding: utf-8 -*-
# psfgen.py

"""
# Copyright (c) 2007-2015, Christoph Gohlke
# Copyright (c) 2007-2015, The Regents of the University of California
# Produced at the Laboratory for Fluorescence Dynamics
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
# * Neither the name of the copyright holders nor the names of any
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

Point spread function calculations for fluorescence microscopy.

:Author:
  `Christoph Gohlke <http://www.lfd.uci.edu/~gohlke/>`_

:Organization:
  Laboratory for Fluorescence Dynamics, University of California, Irvine

:Version: 2015.03.19

Requirements
------------
* `CPython 2.7 or 3.4 <http://www.python.org>`_
* `Numpy 1.9 <http://www.numpy.org>`_
* `Psf.c 2015.03.19 <http://www.lfd.uci.edu/~gohlke/>`_
* `Matplotlib 1.4 <http://www.matplotlib.org>`_  (optional for plotting)

References
----------
(1) Electromagnetic diffraction in optical systems. II. Structure of the
    image field in an aplanatic system.
    B Richards and E Wolf. Proc R Soc Lond A, 253 (1274), 358-379, 1959.
(2) Focal volume optics and experimental artifacts in confocal fluorescence
    correlation spectroscopy.
    S T Hess, W W Webb. Biophys J (83) 2300-17, 2002.
(3) Electromagnetic description of image formation in confocal fluorescence
    microscopy.
    T D Viser, S H Wiersma. J Opt Soc Am A (11) 599-608, 1994.
(4) Photon counting histogram: one-photon excitation.
    B Huang, T D Perroud, R N Zare. Chem Phys Chem (5), 1523-31, 2004.
    Supporting information: Calculation of the observation volume profile.
(5) Gaussian approximations of fluorescence microscope point-spread function
    models.
    B Zhang, J Zerubia, J C Olivo-Marin. Appl. Optics (46) 1819-29, 2007.
(6) The SVI-wiki on 3D microscopy, deconvolution, visualization and analysis.
    http://support.svi.nl/wiki/NyquistRate
(7) Theory of Confocal Microscopy: Resolution and Contrast in Confocal
    Microscopy. http://www.olympusfluoview.com/theory/resolutionintro.html

Examples
--------
# >>> import _psf
# >>> args = dict(shape=(32,32), dims=(4,4), ex_wavelen=488, em_wavelen=520,
# ...             num_aperture=1.2, refr_index=1.333,
# ...             pinhole_radius=0.55, pinhole_shape='round')
# >>> obsvol = psf.PSF(psf.GAUSSIAN | psf.CONFOCAL, **args)
# >>> print("%.5f, %.5f" % obsvol.sigma.ou)
# 2.58832, 1.37059
# >>> obsvol = psf.PSF(psf.ISOTROPIC | psf.CONFOCAL, **args)
# >>> obsvol[0, :3]
# array([ 1.     ,  0.51071,  0.04397])
# >>> # save the image plane to file
# >>> obsvol.slice(0).tofile("_test_slice.bin")
# >>> # save a full 3D PSF volume to file
# >>> obsvol.volume().tofile("_test_volume.bin")

"""



import math
import sys
import threading
import time

import numpy

try:
    from . import _psf
except ImportError:
    raise ImportError(
        "The compiled psf.c extension module could not be found. "
        "Psf.c can be obtained at http://www.lfd.uci.edu/~gohlke/")

from miplib.data.containers.image import Image

__version__ = '2015.03.19'
__docformat__ = 'restructuredtext en'
__all__ = 'PSF', 'Pinhole'

ANISOTROPIC = 1
ISOTROPIC = 2
GAUSSIAN = 4
GAUSSLORENTZ = 8
EXCITATION = 16
EMISSION = 32
WIDEFIELD = 64
CONFOCAL = 128
TWOPHOTON = 256
PARAXIAL = 512


class PsfFromFwhm(object):

    def __init__(self, fwhm, shape=(128, 128), dims=(4., 4.)):
        assert isinstance(fwhm, list)

        if len(fwhm) == 1:
            print ("Only one resolution value given. Assuming the same"
                   " resolution for the axial direction.")
            fwhm = [fwhm, ] * 2

        self.shape = int(shape[0]), int(shape[1])
        self.dims = Dimensions(px=shape, um=(float(dims[0]), float(dims[1])))

        self.spacing = list(x/y for x, y in zip(self.dims.um, self.dims.px))
        self.sigma_px = list(x/(2*math.sqrt(2*math.log(2))*y) for x, y in zip(fwhm, self.spacing))

        self.data = _psf.gaussian2d(self.dims.px, self.sigma_px)

    def xy(self):
        """Return a z slice of the PSF with rotational symmetries applied."""
        data = mirror_symmetry(_psf.zr2zxy(self.data))
        spacing = (self.spacing[1], self.spacing[1])

        center = self.shape[0] - 1
        return Image(data[center], spacing)

    def volume(self):
        """Return a 3D volume of the PSF with all symmetries applied.

        The shape of the returned array is
            (2*self.shape[0]-1, 2*self.shape[1]-1, 2*self.shape[1]-1)

        """
        data = mirror_symmetry(_psf.zr2zxy(self.data))
        spacing = (self.spacing[0], self.spacing[1], self.spacing[1])

        return Image(data, spacing)


class PSF(object):
    """Calculate point spread function of various types.

    Attributes
    ----------
    psftype : int
        A combination of the following properties. Valid combinations are
        listed in PSF.psftype.

        ANISOTROPIC or ISOTROPIC or GAUSSIAN or GAUSSLORENTZ
            Specify calculation model.
        EXCITATION or EMISSION or WIDEFIELD or CONFOCAL or TWOPHOTON
            Specify type of PSF.
        PARAXIAL
            Border case for Gaussian approximations.
    name : str
        A human readable label.
    data : 2D array of floats (C doubles)
        PSF values in z,r space normalized to the value at the origin.
    shape : sequence of int
        Size of the data array in pixel. Default (256, 256)
    dims : Dimension instance
        Dimensions of the data array in px (pixel), um (micrometers),
        ou (optical units), and au (airy units).
    ex_wavelen and em_wavelen : float or None
        Excitation or emission wavelengths in micrometers if applicable.
    num_aperture : float
        Numerical aperture (NA) of the objective. Default 1.2.
    refr_index : float
        Index of refraction of the sample medium. Default 1.333 (water).
    magnification : float
        Total magnification of the optical system. Default 1.0.
    underfilling : float
        Underfilling factor, i.e. the ratio of the radius of the objective
        back aperture to the exp(-2) radius of the excitation laser.
        Default 1.0.
    sigma : Dimension instance or None
        Gaussian sigma parameters in px (pixel), um (micrometers),
        ou (optical units), and au (airy units) if applicable.
    pinhole : Pinhole instance or None
        Pinhole applies to confocal types only.
    expsf, empsf : PSF instance or None
        Excitation or Emission PSF objects if applicable (i.e. when calculated
        intermediately for confocal)

    Notes
    -----
    Calculations of the isotropic PSFs are based on the complex integration
    representation for the diffraction near the image plane proposed by
    Richards and Wolf [1-4].

    Gaussian approximations are calculated according to [5].

    Widefield calculations are used if the pinhole radius is larger than ~8 au.

    Models for polarized excitation or emission light (ANISOTROPIC) and the
    Gaussian-Lorentzian approximation (GAUSSLORENTZ) are not implemented.

    """
    psftypes = {
        ISOTROPIC | EXCITATION: "Excitation, Isotropic",
        ISOTROPIC | EMISSION: "Emission, Isotropic",
        ISOTROPIC | WIDEFIELD: "Widefield, Isotropic",
        ISOTROPIC | CONFOCAL: "Confocal, Isotropic",
        ISOTROPIC | TWOPHOTON: "Two-Photon, Isotropic",
        GAUSSIAN | EXCITATION: "Excitation, Gaussian",
        GAUSSIAN | EMISSION: "Emission, Gaussian",
        GAUSSIAN | WIDEFIELD: "Widefield, Gaussian",  # == Gaussian Emission
        GAUSSIAN | CONFOCAL: "Confocal, Gaussian",
        GAUSSIAN | TWOPHOTON: "Two-Photon, Gaussian",
        GAUSSIAN | EXCITATION | PARAXIAL: "Excitation, Gaussian, Paraxial",
        GAUSSIAN | EMISSION | PARAXIAL: "Emission, Gaussian, Paraxial",
        GAUSSIAN | WIDEFIELD | PARAXIAL: "Widefield, Gaussian, Paraxial",
        GAUSSIAN | CONFOCAL | PARAXIAL: "Confocal, Gaussian, Paraxial",
        GAUSSIAN | TWOPHOTON | PARAXIAL: "Two-Photon, Gaussian, Paraxial",
    }

    def __init__(self, psftype, shape=(256, 256), dims=(4., 4.),
                 ex_wavelen=None, em_wavelen=None, num_aperture=1.2,
                 refr_index=1.333, magnification=1.0, underfilling=1.0,
                 pinhole_radius=None, pinhole_shape='round',
                 expsf=None, empsf=None, name=None):
        """Initialize the PSF object.

        Arguments
        ---------
        psftype, shape, num_aperture, refr_index, magnification, underfilling,
            expsf, and empsf:
            See PSF attributes.
        dims : sequence of float
            Dimensions of the data array in *micrometers*. Default (4., 4.)
        ex_wavelen and em_wavelen : float or None
            Excitation or emission wavelengths in *nanometers* if applicable.
        pinhole_radius : float or None
            Outer radius of the pinhole in *micrometers* in the object space.
            This is the back-projected radius, i.e. the real physical radius
            of the pinhole divided by the magnification of the system.
        pinhole_shape : str
            Either 'round' (default) or 'square'.

        """
        try:
            self.name = self.psftypes[psftype]
            self.psftype = psftype
        except Exception:
            raise ValueError("PSF type is invalid or not supported")

        if name:
            self.name = str(name)

        self.shape = int(shape[0]), int(shape[1])
        self.dims = Dimensions(px=shape, um=(float(dims[0]), float(dims[1])))

        self.ex_wavelen = ex_wavelen / 1e3 if ex_wavelen else None
        self.em_wavelen = em_wavelen / 1e3 if em_wavelen else None
        self.num_aperture = num_aperture
        self.refr_index = refr_index
        self.magnification = magnification
        self.underfilling = underfilling
        self.sigma = None
        self.pinhole = None
        self.expsf = expsf
        self.empsf = empsf

        if (not (psftype & EXCITATION)) and (em_wavelen is None):
            raise ValueError("emission wavelength not specified")

        if (not (psftype & EMISSION)) and (ex_wavelen is None):
            raise ValueError("excitation wavelength not specified")

        if (psftype & CONFOCAL) and (pinhole_radius is None):
            raise ValueError("pinhole radius not specified")

        self.sinalpha = self.num_aperture / self.refr_index
        if self.sinalpha >= 1.0:
            raise ValueError(
                "quotient of the numeric aperture (%.1f) and "
                "refractive index (%.1f) is greater than 1.0 (%.2f)" % (
                    self.num_aperture, self.refr_index, self.sinalpha))

        if psftype & EMISSION:
            au = (1.22 * self.em_wavelen / self.num_aperture)
            ou = zr2uv(self.dims.um, self.em_wavelen, self.sinalpha,
                       self.refr_index, self.magnification)
        else:
            au = (1.22 * self.ex_wavelen / self.num_aperture)
            ou = zr2uv(self.dims.um, self.ex_wavelen, self.sinalpha,
                       self.refr_index, 1.0)
        self.dims.update(ou=ou, au=(self.dims.um[0]/au, self.dims.um[1]/au))

        if pinhole_radius:
            self.pinhole = Pinhole(pinhole_radius, self.dims, pinhole_shape)

        start = time.clock()
        if psftype & GAUSSIAN:
            self.sigma = Dimensions(**self.dims)
            if self.underfilling != 1.0:
                raise NotImplementedError(
                    "underfilling not supported in Gaussian approximation")

            if psftype & EXCITATION or psftype & TWOPHOTON:
                widefield = True
                self.em_wavelen = None
                self.magnification = None
                self.pinh_radius = None
                lex = lem = self.ex_wavelen
                radius = 0.0
            elif psftype & EMISSION or psftype & WIDEFIELD:
                widefield = True
                self.ex_wavelen = None
                self.magnification = None
                lex = lem = self.em_wavelen
                radius = 0.0
            elif psftype & CONFOCAL:
                radius = self.pinhole.radius.um
                if radius > 9.76 * self.ex_wavelen/self.num_aperture:
                    # use widefield approximation for pinholes > 8 AU
                    widefield = True
                    lex = lem = self.ex_wavelen
                else:
                    widefield = False
                    lex = self.ex_wavelen
                    lem = self.em_wavelen
                if self.pinhole.shape != 'round':
                    raise NotImplementedError(
                        "Gaussian approximation only valid for round pinhole")

            paraxial = bool(psftype & PARAXIAL)
            self.sigma.um = _psf.gaussian_sigma(lex, lem, self.num_aperture,
                                                self.refr_index, radius,
                                                widefield, paraxial)
            self.data = _psf.gaussian2d(self.dims.px, self.sigma.px)

        elif psftype & ISOTROPIC:
            if psftype & EXCITATION or psftype & TWOPHOTON:
                self.em_wavelen = None
                self.magnification = None
                self.data = _psf.psf(0, self.shape, self.dims.ou, 1.0,
                                     self.sinalpha, self.underfilling, 1.0, 80)
            elif psftype & EMISSION:
                self.ex_wavelen = None
                self.underfilling = None
                self.data = _psf.psf(1, self.shape, self.dims.ou,
                                     self.magnification, self.sinalpha,
                                     1.0, 1.0, 80)
            elif psftype & CONFOCAL or psftype & WIDEFIELD:
                if em_wavelen < ex_wavelen:
                    raise ValueError("Excitation > Emission wavelength")
                # start threads to calculate excitation and emission PSF
                threads = []
                if not (self.expsf and
                        self.expsf.psftype == ISOTROPIC | EXCITATION):
                    threads.append((
                        "expsf",
                        PSFthread(ISOTROPIC | EXCITATION,
                                  shape, dims, ex_wavelen, None, num_aperture,
                                  refr_index, 1.0, underfilling)))
                if not (self.empsf and
                        self.empsf.psftype == ISOTROPIC | EMISSION):
                    threads.append((
                        "empsf",
                        PSFthread(ISOTROPIC | EMISSION,
                                  shape, dims, None, em_wavelen, num_aperture,
                                  refr_index, magnification, 1.0)))
                for a, t in threads:
                    t.start()
                for a, t in threads:
                    t.join()
                    setattr(self, a, t.psf)
                if not (self.expsf.iscompatible(self.empsf)):
                    raise ValueError(
                        "Excitation and Emission PSF not compatible")
                if psftype & WIDEFIELD or (self.pinhole.radius.um > 9.76 *
                                           self.ex_wavelen / self.num_aperture
                                           ):
                    # use widefield approximation for pinholes > 8 AU
                    self.data = _psf.obsvol(self.expsf.data, self.empsf.data)
                else:
                    self.data = _psf.obsvol(self.expsf.data, self.empsf.data,
                                            self.pinhole.kernel())

        if psftype & TWOPHOTON:
            self.data *= self.data
        self.time = float(time.clock()-start) * 1e3

    def __getitem__(self, key):
        """Return value of data array at position."""
        return self.data[key]

    def __str__(self):
        """Return properties of PSF object as string."""
        s = [self.name]
        s.append("  Shape: (%i, %i) pixel" % self.dims.px)
        s.append("  Dimensions: %s" % self.dims.format(
            ["um", "ou", "au"], ["%.2f", "%.2f", "%.2f"]))
        if self.ex_wavelen:
            s.append(
                "  Excitation Wavelength: %.1f nm" % (self.ex_wavelen * 1e3))
        if self.em_wavelen:
            s.append(
                "  Emission Wavelength: %.1f nm" % (self.em_wavelen * 1e3))
        s.append("  Numeric Aperture: %.2f" % self.num_aperture)
        s.append("  Refractive Index: %.2f" % self.refr_index)
        s.append("  Half Cone Angle: %.2f deg" % math.degrees(
            math.asin(self.sinalpha)))
        if self.magnification:
            s.append("  Magnification: %.2f" % self.magnification)
        if self.underfilling:
            s.append("  Underfilling: %.2f" % self.underfilling)
        if self.pinhole:
            s.append("  Pinhole Radius: %s" % self.pinhole.radius.format(
                ["um", "ou", "au", "px"], ["%.3f", "%.3f", "%.4f", "%.2f"]))
        if self.sigma:
            s.append("  Gauss Sigma: %s" % self.sigma.format(
                ["um", "ou", "au", "px"], ["%.3f", "%.3f", "%.3f", "%.2f"]))
        s.append("  Computing Time: %.2f ms\n" % self.time)
        return "\n".join(s)

    def iscompatible(self, other):
        """Return True if objects match dimensions and optical properties."""
        return ((self.dims.px[0] == other.dims.px[0])
                and (self.dims.px[1] == other.dims.px[1])
                and (self.dims.um[0] == other.dims.um[0])
                and (self.dims.um[1] == other.dims.um[1])
                and (self.num_aperture == other.num_aperture)
                and (self.refr_index == other.refr_index))

    def slice(self, key=slice(None)):
        """Return a z slice of the PSF with rotational symmetries applied."""
        return _psf.zr2zxy(self.data[key])

    def volume(self):
        """Return a 3D volume of the PSF with all symmetries applied.

        The shape of the returned array is
            (2*self.shape[0]-1, 2*self.shape[1]-1, 2*self.shape[1]-1)

        """
        return mirror_symmetry(_psf.zr2zxy(self.data))

    def imshow(self, subplot=111, **kwargs):
        """Log-plot PSF image using matplotlib.pyplot. Return plot axis."""
        title = kwargs.get("title", self.name)
        aspect = self.shape[1]/self.shape[0] * self.dims.um[0]/self.dims.um[1]
        kwargs.update(dict(data=self.data, title=title, subplot=subplot,
                           aspect=aspect))
        return imshow(**kwargs)

    def sted_correction(self, phi=3.22, sigma=6.0):
        """
        In STED PSF the regular confocal PSF gets modified by the depletion donut.
        This function does the corrections as eplained in:

        Zanella, R. et al., 2013. Towards real-time image deconvolution: application to
        confocal and STED microscopy. Scientific reports, 3, p.2523.
        Available at: http://www.nature.com/srep/2013/130828/srep02523/full/srep02523.htm.

        :param phi:   depletion gradient / um
        :param sigma: saturation factor Isted/Isat

        """
        phi *= self.dims.um[0]/self.dims.px[0]
        r_s = numpy.arange(0, self.data.shape[1], 1)
        multiplier = 1.0/(1+4*phi**2*sigma*r_s**2)

        self.data *= multiplier[None, :]



class PSFthread(threading.Thread):
    """Calculate point spread function in a thread."""

    def __init__(self, *args, **kwargs):
        threading.Thread.__init__(self)
        self.args = args
        self.kwargs = kwargs
        self.psf = None

    def run(self):
        self.psf = PSF(*self.args, **self.kwargs)


class Pinhole(object):
    """Pinhole object for confocal microscopy.

    Attributes
    ----------
    radius : Dimension instance
        Dimensions of the outer pinhole radius in px (pixel), um (micrometers),
        ou (optical units), and au (airy units).
    shape : str
        Shape of the pinhole. Either 'round' or 'square'.

    Examples
    --------
    >>> ph = Pinhole(0.1, dict(px=16, um=1.0), 'round')
    >>> print("%.5f" % ph.radius.px)
    1.60000
    >>> ph.kernel()
    array([[ 1.     ,  1.6    ,  0.6    ],
           [ 0.8    ,  1.18579,  0.36393],
           [ 0.3    ,  0.36393,  0.     ]])

    """
    shapes = {'round': 0, 'square': 4}

    def __init__(self, radius, dimensions, shape='round'):
        """Initialize the pinhole object.

        Arguments
        ---------
        radius : float
            Outer pinhole radius in micrometers in object space.
        dimensions : dict
            Dimensions of the object space in "px" (pixel), "um" (micrometers),
            "ou" (optical units), and "au" (airy units).
        shape : str
            Shape of the pinhole. Either 'round' (default) or 'square'.

        """
        self._corners = self.shapes[shape]
        self.shape = shape
        try:
            dimensions = dict((k, v[1]) for k, v in list(dimensions.items()))
        except TypeError:
            pass
        self.radius = Dimensions(**dimensions)
        self.radius.um = float(radius)
        self._kernel = None

    def __str__(self):
        s = [self.shape, "  Radius: %s" % str(self.radius)]
        return "\n".join(s)

    def kernel(self):
        """Return convolution kernel for integration over the pinhole."""
        if self._kernel is None:
            self._kernel = _psf.pinhole_kernel(self.radius.px, self._corners)
        return self._kernel


class Dimensions(dict):
    """Store dimensions in various units and perform linear conversions.

    Examples
    --------
    >>> dim = Dimensions(px=100, um=2)
    >>> dim(50, "px", "um")
    1.0
    >>> dim.px, dim.um
    (100, 2)
    >>> dim.px = 50
    >>> dim.um
    1.0
    >>> dim.format(("um", "px"), ("%.2f", "%.1f"))
    '1.00 um, 50.0 px'
    >>> dim = Dimensions(px=(100, 200), um=(2, 8))
    >>> dim((50, 50), "px", "um")
    (1.0, 2.0)
    >>> dim.ou = (1, 2)
    >>> dim.px
    (100, 200)
    >>> dim["px"] = (50, 100)
    >>> dim.ou
    (0.5, 1.0)

    """
    __slots__ = []

    def __call__(self, value, unit, newunit):
        """Return value given in unit in another unit."""
        dim = self[unit]
        new = self[newunit]
        try:
            return value * (new/dim)
        except TypeError:
            return tuple(v*(o/u) for v, u, o in zip(value, dim, new))

    def __setitem__(self, unit, value):
        """Add a dimension or rescale all dimensions to new value."""
        try:
            dim = self[unit]
        except KeyError:
            dict.__setitem__(self, unit, value)
        else:
            try:
                scale = value / dim
                for k, v in list(self.items()):
                    dict.__setitem__(self, k, v * scale)
            except TypeError:
                scale = tuple(v/d for v, d in zip(value, dim))
                for k, v in list(self.items()):
                    dict.__setitem__(
                        self, k, tuple(v*s for v, s in zip(self[k], scale)))

    def __getattr__(self, unit):
        """Return value of unit."""
        return self[unit]

    def __setattr__(self, unit, value):
        """Add a dimension or rescale all dimensions to new value."""
        self.__setitem__(unit, value)

    def format(self, keys, formatstr):
        """Return formatted string."""
        s = []
        try:
            for k, f in zip(keys, formatstr):
                s.append("%s %s" % (f % self[k], k))
        except TypeError:
            for k, f in zip(keys, formatstr):
                v = self[k]
                t = []
                for i in v:
                    t.append(f % i)
                s.append("(%s) %s" % (", ".join(t), k))
        return ", ".join(s)


def uv2zr(uv, wavelength, sinalpha, refr_index, magnification=1.0):
    """Return z,r in units of the wavelength from u,v given in optical units.

    For excitation, magnification should be 1.

    Examples
    --------
    >>> numpy.allclose(uv2zr((1, 1), 488, 0.9, 1.33),
    ...               (72.094692498695736, 64.885223248826165))
    True

    """
    a = wavelength / (2.0 * math.pi * sinalpha * refr_index * magnification)
    b = a / (sinalpha * magnification)
    return uv[0]*b, uv[1]*a


def zr2uv(zr, wavelength, sinalpha, refr_index, magnification=1.0):
    """Return u,v in optical units from z,r given in units of the wavelength.

    For excitation, magnification should be 1.

    Examples
    --------
    >>> numpy.allclose(zr2uv((1e3, 1e3), 488, 0.9, 1.33),
    ...               (13.870646580788051, 15.411829534208946))
    True

    """
    a = (2.0 * math.pi * refr_index * sinalpha * magnification) / wavelength
    b = a * sinalpha * magnification
    return zr[0]*b, zr[1]*a


def mirror_symmetry(data):
    """Apply mirror symmetry along one face in each dimension.

    The input array can be 1, 2 or 3-dimensional.

    The shape of the returned array is 2*data.shape-1 in each dimension.

    Examples
    --------
    >>> mirror_symmetry([0, 1])
    array([ 1.,  0.,  1.])
    >>> mirror_symmetry([[0, 1],[0, 1]])
    array([[ 1.,  0.,  1.],
           [ 1.,  0.,  1.],
           [ 1.,  0.,  1.]])
    >>> mirror_symmetry([[[0, 1],[0, 1]], [[0, 1],[0, 1]], [[0, 1],[0, 1]]])[0]
    array([[ 1.,  0.,  1.],
           [ 1.,  0.,  1.],
           [ 1.,  0.,  1.]])

    """
    data = numpy.array(data)
    result = numpy.empty([2*i-1 for i in data.shape], numpy.float64)
    if data.ndim == 1:
        x = data.shape[0] - 1
        result[x:] = data
        result[:x] = data[-1:0:-1]
    elif data.ndim == 2:
        x, y = (i-1 for i in data.shape)
        result[x:, y:] = data
        result[:x, y:] = data[-1:0:-1, :]
        result[:, :y] = result[:, -1:y:-1]
    elif data.ndim == 3:
        x, y, z = (i - 1 for i in data.shape)
        result[x:, y:, z:] = data
        result[:x, y:, z:] = data[-1:0:-1, :, :]
        result[:, :y, z:] = result[:, -1:y:-1, z:]
        result[:, :, :z] = result[:, :, -1:z:-1]
    else:
        raise NotImplementedError("%i-dimensional arrays not supported" %
                                  data.ndim)
    return result


def imshow(subplot, data, title=None, sharex=None, sharey=None,
           vmin=-2.5, vmax=0.0, cmap=None, interpolation='bilinear', **kwargs):
    """Log-plot image using matplotlib.pyplot. Return plot axis and plot.

    Mirror symmetry is applied along the x and y axes.

    Requires pyplot already imported ``from matplotlib import pyplot``.

    """
    pyplot = sys.modules['matplotlib.pyplot']

    ax = pyplot.subplot(subplot, sharex=sharex, sharey=sharey, axisbg='k')
    if title:
        pyplot.title(title)
    if cmap is None:
        cmap = pyplot.cm.cubehelix  # coolwarm
    # workaround: set alpha for i_bad
    cmap._init()
    cmap._lut[-1, -1] = 1.0

    im = pyplot.imshow(mirror_symmetry(numpy.log10(data)),
                       vmin=vmin, vmax=vmax, cmap=cmap,
                       interpolation=interpolation, **kwargs)
    pyplot.axis('off')
    return ax, im


if __name__ == "__main__":
    import doctest
    numpy.set_printoptions(suppress=True, precision=5)
    doctest.testmod()